"""An example to run of the minitaur gym environment with sine gaits.

"""

import csv
import math
import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(os.path.dirname(currentdir)))
print("parentdir=", parentdir)
os.sys.path.insert(0, parentdir)

import argparse
import numpy as np
import tf.compat.v1 as tf
from pybullet_envs.minitaur.envs import minitaur_gym_env
import time

#FLAGS = flags.FLAGS
#flags.DEFINE_enum(
#    "example_name", "sine", ["sine", "reset", "stand", "overheat"],
#    "The name of the example: sine, reset, stand, or overheat.")
#flags.DEFINE_string("output_filename", None, "The name of the output CSV file."
#                    "Each line in the CSV file will contain the action, the "
#                    "motor position, speed and torques.")
#flags.DEFINE_string("log_path", None, "The directory to write the log file.")


def WriteToCSV(filename, actions_and_observations):
  """Write simulation data to file.

  Save actions and observed angles, angular velocities and torques for data
  analysis.

  Args:
    filename: The file to write. Can be locally or on CNS.
    actions_and_observations: the interested simulation quantities to save.
  """
  with tf.gfile.Open(filename, "wb") as csvfile:
    csv_writer = csv.writer(csvfile, delimiter=",")
    for row in actions_and_observations:
      csv_writer.writerow(row)


def ResetPoseExample(log_path=None):
  """An example that the minitaur stands still using the reset pose."""
  steps = 10000
  environment = minitaur_gym_env.MinitaurGymEnv(
      urdf_version=minitaur_gym_env.DERPY_V0_URDF_VERSION,
      render=True,
      leg_model_enabled=False,
      motor_velocity_limit=np.inf,
      pd_control_enabled=True,
      accurate_motor_model_enabled=True,
      motor_overheat_protection=True,
      hard_reset=False,
      log_path=log_path)
  action = [math.pi / 2] * 8
  for _ in range(steps):
    _, _, done, _ = environment.step(action)
    time.sleep(1. / 100.)
    if done:
      break


def MotorOverheatExample(log_path=None):
  """An example of minitaur motor overheat protection is triggered.

  The minitaur is leaning forward and the motors are getting obove threshold
  torques. The overheat protection will be triggered in ~1 sec.

  Args:
    log_path: The directory that the log files are written to. If log_path is
      None, no logs will be written.
  """

  environment = minitaur_gym_env.MinitaurGymEnv(
      urdf_version=minitaur_gym_env.DERPY_V0_URDF_VERSION,
      render=True,
      leg_model_enabled=False,
      motor_velocity_limit=np.inf,
      motor_overheat_protection=True,
      accurate_motor_model_enabled=True,
      motor_kp=1.20,
      motor_kd=0.00,
      on_rack=False,
      log_path=log_path)

  action = [2.0] * 8
  for i in range(8):
    action[i] = 2.0 - 0.5 * (-1 if i % 2 == 0 else 1) * (-1 if i < 4 else 1)

  steps = 500
  actions_and_observations = []
  for step_counter in range(steps):
    # Matches the internal timestep.
    time_step = 0.01
    t = step_counter * time_step
    current_row = [t]
    current_row.extend(action)

    observation, _, _, _ = environment.step(action)
    current_row.extend(observation.tolist())
    actions_and_observations.append(current_row)
    time.sleep(1. / 100.)

  if FLAGS.output_filename is not None:
    WriteToCSV(FLAGS.output_filename, actions_and_observations)


def SineStandExample(log_path=None):
  """An example of minitaur standing and squatting on the floor.

  To validate the accurate motor model we command the robot and sit and stand up
  periodically in both simulation and experiment. We compare the measured motor
  trajectories, torques and gains. The results are at:
    https://colab.corp.google.com/v2/notebook#fileId=0BxTIAnWh1hb_ZnkyYWtNQ1RYdkU&scrollTo=ZGFMl84kKqRx

  Args:
    log_path: The directory that the log files are written to. If log_path is
      None, no logs will be written.
  """
  environment = minitaur_gym_env.MinitaurGymEnv(
      urdf_version=minitaur_gym_env.RAINBOW_DASH_V0_URDF_VERSION,
      render=True,
      leg_model_enabled=False,
      motor_velocity_limit=np.inf,
      motor_overheat_protection=True,
      accurate_motor_model_enabled=True,
      motor_kp=1.20,
      motor_kd=0.02,
      on_rack=False,
      log_path=log_path)
  steps = 1000
  amplitude = 0.5
  speed = 3

  actions_and_observations = []

  for step_counter in range(steps):
    # Matches the internal timestep.
    time_step = 0.01
    t = step_counter * time_step
    current_row = [t]

    action = [math.sin(speed * t) * amplitude + math.pi / 2] * 8
    current_row.extend(action)

    observation, _, _, _ = environment.step(action)
    current_row.extend(observation.tolist())
    actions_and_observations.append(current_row)
    time.sleep(1. / 100.)

  if FLAGS.output_filename is not None:
    WriteToCSV(FLAGS.output_filename, actions_and_observations)


def SinePolicyExample(log_path=None):
  """An example of minitaur walking with a sine gait.

  Args:
    log_path: The directory that the log files are written to. If log_path is
      None, no logs will be written.
  """
  environment = minitaur_gym_env.MinitaurGymEnv(
      urdf_version=minitaur_gym_env.DERPY_V0_URDF_VERSION,
      render=True,
      motor_velocity_limit=np.inf,
      pd_control_enabled=True,
      hard_reset=False,
      on_rack=False,
      log_path=log_path)
  sum_reward = 0
  steps = 20000
  amplitude_1_bound = 0.5
  amplitude_2_bound = 0.5
  speed = 40

  for step_counter in range(steps):
    time_step = 0.01
    t = step_counter * time_step

    amplitude1 = amplitude_1_bound
    amplitude2 = amplitude_2_bound
    steering_amplitude = 0
    if t < 10:
      steering_amplitude = 0.5
    elif t < 20:
      steering_amplitude = -0.5
    else:
      steering_amplitude = 0

    # Applying asymmetrical sine gaits to different legs can steer the minitaur.
    a1 = math.sin(t * speed) * (amplitude1 + steering_amplitude)
    a2 = math.sin(t * speed + math.pi) * (amplitude1 - steering_amplitude)
    a3 = math.sin(t * speed) * amplitude2
    a4 = math.sin(t * speed + math.pi) * amplitude2
    action = [a1, a2, a2, a1, a3, a4, a4, a3]
    _, reward, done, _ = environment.step(action)
    time.sleep(1. / 100.)

    sum_reward += reward
    if done:
      tf.logging.info("Return is {}".format(sum_reward))
      environment.reset()


def main():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--env',
                      help='environment ID (0==sine, 1==stand, 2=reset, 3=overheat)',
                      type=int,
                      default=0)
  args = parser.parse_args()
  print("--env=" + str(args.env))

  if (args.env == 0):
    SinePolicyExample()
  if (args.env == 1):
    SineStandExample()
  if (args.env == 2):
    ResetPoseExample()
  if (args.env == 3):
    MotorOverheatExample()


if __name__ == '__main__':
  main()



import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

from pybullet_utils import bullet_client
import panda_sim


import time

useGUI = False
timeStep = 1./60.

# Importing the libraries
import os
import time
import multiprocessing as mp
from multiprocessing import Process, Pipe

pandaEndEffectorIndex = 11 #8
pandaNumDofs = 7


_RESET = 1
_CLOSE = 2
_EXPLORE = 3


def ExploreWorker(rank, num_processes, childPipe, args):
  print("hi:",rank, " out of ", num_processes)  
  import pybullet as op1
  import pybullet_data as pd
  logName=""
  p1=0
  n = 0
  space = 2 
  simulations=[]
  sims_per_worker = 10
  
  offsetY = rank*space
  while True:
    n += 1
    try:
      # Only block for short times to have keyboard exceptions be raised.
      if not childPipe.poll(0.0001):
        continue
      message, payload = childPipe.recv()
    except (EOFError, KeyboardInterrupt):
      break
    if message == _RESET:
      if (useGUI):
        p1 = bullet_client.BulletClient(op1.GUI)
      else:
        p1 = bullet_client.BulletClient(op1.DIRECT)
      p1.setTimeStep(timeStep)
      
      p1.setPhysicsEngineParameter(numSolverIterations=8)
      p1.setPhysicsEngineParameter(minimumSolverIslandSize=100)
      p1.configureDebugVisualizer(p1.COV_ENABLE_Y_AXIS_UP,1)
      p1.configureDebugVisualizer(p1.COV_ENABLE_RENDERING,0)
      p1.setAdditionalSearchPath(pd.getDataPath())
      p1.setGravity(0,-9.8,0)
      logName = str("batchsim")+str(rank)
      for j in range (3):
        offsetX = 0#-sims_per_worker/2.0*space
        for i in range (sims_per_worker):
          offset=[offsetX,0, offsetY]
          sim = panda_sim.PandaSim(p1, offset)
          simulations.append(sim)
          offsetX += space 
        offsetY += space 
      childPipe.send(["reset ok"])
      p1.configureDebugVisualizer(p1.COV_ENABLE_RENDERING,1)
      for i in range (100):
        p1.stepSimulation()
      
      logId = p1.startStateLogging(op1.STATE_LOGGING_PROFILE_TIMINGS,logName)
      continue
    if message == _EXPLORE:
      sum_rewards=rank
      
      if useGUI:
        numSteps = int(20000)
      else:
        numSteps = int(5)
      for i in range (numSteps):
        for s in simulations:
          s.step()
        p1.stepSimulation()
      #print("logId=",logId)
      #print("numSteps=",numSteps)

      childPipe.send([sum_rewards])
      continue
    if message == _CLOSE:
      p1.stopStateLogging(logId)
      childPipe.send(["close ok"])
      break
  childPipe.close()
  

if __name__ == "__main__":
  mp.freeze_support()
  if useGUI:
    num_processes = 1
  else:
    num_processes = 12
  processes = []
  args=[0]*num_processes
  
  childPipes = []
  parentPipes = []

  for pr in range(num_processes):
    parentPipe, childPipe = Pipe()
    parentPipes.append(parentPipe)
    childPipes.append(childPipe)

  for rank in range(num_processes):
    p = mp.Process(target=ExploreWorker, args=(rank, num_processes, childPipes[rank],  args))
    p.start()
    processes.append(p)
      
  
  for parentPipe in parentPipes:
    parentPipe.send([_RESET, "blaat"])
  
  positive_rewards = [0]*num_processes
  for k in range(num_processes):
    #print("reset msg=",parentPipes[k].recv()[0])
    msg = parentPipes[k].recv()[0]
  
  for parentPipe in parentPipes:
    parentPipe.send([_EXPLORE, "blaat"])
  
  positive_rewards = [0]*num_processes
  for k in range(num_processes):
    positive_rewards[k] = parentPipes[k].recv()[0]
    #print("positive_rewards=",positive_rewards[k])

  
  for parentPipe in parentPipes:
    parentPipe.send([_EXPLORE, "blaat"])
  
  positive_rewards = [0]*num_processes
  for k in range(num_processes):
    positive_rewards[k] = parentPipes[k].recv()[0]
    #print("positive_rewards=",positive_rewards[k])
    msg = positive_rewards[k]

  for parentPipe in parentPipes:
    parentPipe.send([_EXPLORE, "blaat"])
  
  positive_rewards = [0]*num_processes
  for k in range(num_processes):
    positive_rewards[k] = parentPipes[k].recv()[0]
    #print("positive_rewards=",positive_rewards[k])

  
  for parentPipe in parentPipes:
    parentPipe.send([_CLOSE, "pay2"])
  
  for p in processes:
    p.join()
  
  #now we merge the separate json files into a single one
  fnameout = 'batchsim.json'
  count = 0
  outfile = open(fnameout, "w+")
  outfile.writelines(["{\"traceEvents\":[\n"])
  numFiles = num_processes
  for num in range(numFiles):
    print("num=",num)
    fname = 'batchsim%d_0.json' % (num)
    with open(fname) as infile:
        for line in infile:
          if "pid" in line:
            line = line.replace('\"pid\":1', '\"pid\":'+str(num))
            if num < (numFiles-1) and not "{}}," in line:
              line = line.replace('{}}', '{}},')
              print("line[",count,"]=",line)
            outfile.write(line)
          count += 1
  print ("count=",count)
  outfile.writelines(["],\n"])
  outfile.writelines(["\"displayTimeUnit\": \"ns\"}\n"])
  outfile.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

#python proto2csv.py --proto_file=/tmp/logs/minitaur_log_2019-01-27-12-59-31 --csv_file=/tmp/logs/out.csv
#each line in csv contains: angle, velocity, action, torque

import tf.compat.v1 as tf
import argparse
import numpy
from pybullet_envs.minitaur.envs import minitaur_logging

flags = tf.app.flags
FLAGS = tf.app.flags.FLAGS

flags.DEFINE_string("proto_file", "logs", "path to protobuf input file")
flags.DEFINE_string("csv_file", "file.csv", "poth to csv output file")


def main(argv):
  del argv

  logging = minitaur_logging.MinitaurLogging()
  episode = logging.restore_episode(FLAGS.proto_file)
  #print(dir (episode))
  #print("episode=",episode)
  fields = episode.ListFields()

  recs = []

  for rec in fields[0][1]:
    #print(rec.time)
    for motorState in rec.motor_states:
      #print("motorState.angle=",motorState.angle)
      #print("motorState.velocity=",motorState.velocity)
      #print("motorState.action=",motorState.action)
      #print("motorState.torque=",motorState.torque)
      recs.append([motorState.angle, motorState.velocity, motorState.action, motorState.torque])

  a = numpy.array(recs)
  numpy.savetxt(FLAGS.csv_file, a, delimiter=",")


if __name__ == "__main__":
  tf.app.run(main)



# Lint as: python3
"""A torque based stance controller framework."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import logging
from typing import Any, Sequence, Tuple

import gin
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import gait_generator as gait_generator_lib
from pybullet_envs.minitaur.agents.baseline_controller import leg_controller
#from pybullet_envs.minitaur.agents.baseline_controller.convex_mpc.python import convex_mpc
#from google3.util.task.python import error

try:
  import mpc_osqp as convex_mpc  # pytype: disable=import-error
except:  #pylint: disable=W0702
  print("You need to install motion_imitation")
  print("or use pip3 install motion_imitation --user")
  print("see also https://github.com/google-research/motion_imitation")
  import sys
  sys.exit()
  

_FORCE_DIMENSION = 3
# The QP weights in the convex MPC formulation. See the MIT paper for details:
#   https://ieeexplore.ieee.org/document/8594448/
# Intuitively, this is the weights of each state dimension when tracking a
# desired CoM trajectory. The full CoM state is represented by
# (roll_pitch_yaw, position, angular_velocity, velocity, gravity_place_holder).
_MPC_WEIGHTS = (5, 5, 0.2, 0, 0, 10, 0.5, 0.5, 0.2, 0.2, 0.2, 0.1, 0)
_PLANNING_HORIZON_STEPS = 10
_PLANNING_TIMESTEP = 0.025
#_MPC_CONSTRUCTOR = functools.partial(
#    convex_mpc.ConvexMpc, qp_solver_name=convex_mpc.QPSolverName.QPOASES)


@gin.configurable
class TorqueStanceLegController(leg_controller.LegController):
  """A torque based stance leg controller framework.

  Takes in high level parameters like walking speed and turning speed, and
  generates necessary the torques for stance legs.
  """

  def __init__(
      self,
      robot: Any,
      gait_generator: Any,
      state_estimator: Any,
      desired_speed: Tuple[float] = (0, 0),
      desired_twisting_speed: float = 0,
      desired_roll_pitch: Tuple[float] = (0, 0),
      desired_body_height: float = 0.45,
      body_mass: float = 220 / 9.8,
      body_inertia: Tuple[float] = (0.183375, 0, 0, 0, 0.6267, 0, 0, 0,
                                    0.636175),
      num_legs: int = 4,
      friction_coeffs: Sequence[float] = (0.5, 0.5, 0.5, 0.5),
      qp_weights: Sequence[float] = _MPC_WEIGHTS,
      planning_horizon: int = _PLANNING_HORIZON_STEPS,
      planning_timestep: int = _PLANNING_TIMESTEP,
  ):
    """Initializes the class.

    Tracks the desired position/velocity of the robot by computing proper joint
    torques using MPC module.

    Args:
      robot: A robot instance.
      gait_generator: Used to query the locomotion phase and leg states.
      state_estimator: Estimate the robot states (e.g. CoM velocity).
      desired_speed: desired CoM speed in x-y plane.
      desired_twisting_speed: desired CoM rotating speed in z direction.
      desired_roll_pitch: desired CoM roll and pitch.
      desired_body_height: The standing height of the robot.
      body_mass: The total mass of the robot.
      body_inertia: The inertia matrix in the body principle frame. We assume
        the body principle coordinate frame has x-forward and z-up.
      num_legs: The number of legs used for force planning.
      friction_coeffs: The friction coeffs on the contact surfaces.
      qp_weights: The weights used in solving the QP problem.
      planning_horizon: Number of steps to roll-out in the QP formulation.
      planning_timestep: Timestep between each step in the QP formulation.
    """

    self._robot = robot
    self._gait_generator = gait_generator
    self._state_estimator = state_estimator
    self._desired_speed = desired_speed
    self._desired_twisting_speed = desired_twisting_speed
    self._desired_roll_pitch = desired_roll_pitch
    self._desired_body_height = desired_body_height
    self._body_mass = body_mass
    self._num_legs = num_legs
    self._friction_coeffs = np.array(friction_coeffs)
    self._qp_solver_fail = False
    self._com_estimate_leg_indices = None
    qp_solver = convex_mpc.QPOASES #convex_mpc.OSQP #
    
    body_inertia_list = list(body_inertia)
    weights_list = list(qp_weights)

    self._mpc = convex_mpc.ConvexMpc(
        body_mass,
        body_inertia_list,
        self._num_legs,
        planning_horizon,
        planning_timestep,
        weights_list,
        1e-6,
        qp_solver
    )
    

  def reset(self, current_time):
    del current_time
    self._qp_solver_fail = False
    self._com_estimate_leg_indices = None

  def update(self, current_time):
    del current_time

  def get_action(self):
    """Computes the torque for stance legs."""
    desired_com_position = np.array((0., 0., self._desired_body_height),
                                    dtype=np.float64)
    desired_com_velocity = np.array(
        (self.desired_speed[0], self.desired_speed[1], 0.), dtype=np.float64)
    desired_com_roll_pitch_yaw = np.array(
        (self.desired_roll_pitch[0], self.desired_roll_pitch[1], 0.),
        dtype=np.float64)
    desired_com_angular_velocity = np.array(
        (0., 0., self.desired_twisting_speed), dtype=np.float64)
    foot_contact_state = np.array(
        [(leg_state == gait_generator_lib.LegState.STANCE or
          leg_state == gait_generator_lib.LegState.EARLY_CONTACT)
         for leg_state in self._gait_generator.desired_leg_state],
        dtype=np.int32)

    # We use the body yaw aligned world frame for MPC computation.
    com_roll_pitch_yaw = np.array(
        self._robot.base_roll_pitch_yaw, dtype=np.float64)
    com_roll_pitch_yaw[2] = 0
    #try:
    estimated_com_position = np.array(())
    if hasattr(self._state_estimator, "estimated_com_height"):
        estimated_com_position = np.array(
            (0, 0, self._state_estimator.estimated_com_height))
    try:
      predicted_contact_forces = self._mpc.compute_contact_forces(
        estimated_com_position,  #com_position
        np.asarray(self._state_estimator.com_velocity_body_yaw_aligned_frame,
                   dtype=np.float64),  #com_velocity
        np.array(com_roll_pitch_yaw, dtype=np.float64),  #com_roll_pitch_yaw
        # Angular velocity in the yaw aligned world frame is actually different
        # from rpy rate. We use it here as a simple approximation.
        np.asarray(self._robot.base_roll_pitch_yaw_rate,
                   dtype=np.float64),  #com_angular_velocity
        foot_contact_state,  #foot_contact_states
        np.array(self._robot.foot_positions(
              position_in_world_frame=False).flatten(),
                 dtype=np.float64),  #foot_positions_base_frame
        self._friction_coeffs,  #foot_friction_coeffs
        desired_com_position,  #desired_com_position
        desired_com_velocity,  #desired_com_velocity
        desired_com_roll_pitch_yaw,  #desired_com_roll_pitch_yaw
        desired_com_angular_velocity  #desired_com_angular_velocity
    )
    except:# error.StatusNotOk as e:
      logging.error("Error in Torque Stance Leg")#e.message)
      self._qp_solver_fail = True
      predicted_contact_forces = np.zeros(self._num_legs * _FORCE_DIMENSION)

    contact_forces = {}
    for i in range(self._num_legs):
      contact_forces[i] = np.array(
          predicted_contact_forces[i * _FORCE_DIMENSION:(i + 1) *
                                   _FORCE_DIMENSION])

    _, kds = self._robot.motor_model.get_motor_gains()
    action = {}
    for leg_id, force in contact_forces.items():
      motor_torques = self._robot.map_contact_force_to_joint_torques(
          leg_id, force)
      for joint_id, torque in motor_torques.items():
        action[joint_id] = (0, 0, 0, kds[joint_id], torque)
    return action

  @property
  def qp_solver_fail(self):
    return self._qp_solver_fail

  @property
  def desired_speed(self):
    return self._desired_speed

  @desired_speed.setter
  def desired_speed(self, speed):
    self._desired_speed = speed

  @property
  def desired_twisting_speed(self):
    return self._desired_twisting_speed

  @desired_twisting_speed.setter
  def desired_twisting_speed(self, twisting_speed):
    self._desired_twisting_speed = twisting_speed

  @property
  def desired_roll_pitch(self):
    return self._desired_roll_pitch

  @desired_roll_pitch.setter
  def desired_roll_pitch(self, roll_pitch):
    self._desired_roll_pitch = roll_pitch

  @property
  def desired_body_height(self):
    return self._desired_body_height

  @desired_body_height.setter
  def desired_body_height(self, body_height):
    self._desired_body_height = body_height



import pybullet as p
import time
import pybullet_data

cid = p.connect(p.SHARED_MEMORY)
if (cid < 0):
  p.connect(p.GUI)

p.setAdditionalSearchPath(pybullet_data.getDataPath())
q = p.loadURDF("quadruped/quadruped.urdf", useFixedBase=True)
rollId = p.addUserDebugParameter("roll", -1.5, 1.5, 0)
pitchId = p.addUserDebugParameter("pitch", -1.5, 1.5, 0)
yawId = p.addUserDebugParameter("yaw", -1.5, 1.5, 0)
fwdxId = p.addUserDebugParameter("fwd_x", -1, 1, 0)
fwdyId = p.addUserDebugParameter("fwd_y", -1, 1, 0)
fwdzId = p.addUserDebugParameter("fwd_z", -1, 1, 0)

while True:
  roll = p.readUserDebugParameter(rollId)
  pitch = p.readUserDebugParameter(pitchId)
  yaw = p.readUserDebugParameter(yawId)
  x = p.readUserDebugParameter(fwdxId)
  y = p.readUserDebugParameter(fwdyId)
  z = p.readUserDebugParameter(fwdzId)

  orn = p.getQuaternionFromEuler([roll, pitch, yaw])
  p.resetBasePositionAndOrientation(q, [x, y, z], orn)
  #p.stepSimulation()#not really necessary for this demo, no physics used



"""A moving-window filter for smoothing the signals within certain time interval."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import gin
import numpy as np


@gin.configurable
class TimeBasedMovingWindowFilter:
  """A moving-window filter for smoothing the signals within certain time interval."""

  def __init__(
      self,
      filter_window: float = 0.1,
  ):
    """Initializes the class.

    Args:
      filter_window: The filtering window (in time) used to smooth the input
        signal.
    """
    self._filter_window = filter_window
    self.reset()

  def reset(self):
    self._timestamp_buffer = []
    self._value_buffer = []

  def calculate_average(self, new_value, timestamp):
    """Compute the filtered signals based on the time-based moving window."""
    self._timestamp_buffer.append(timestamp)
    self._value_buffer.append(new_value)

    while len(self._value_buffer) > 1:
      if self._timestamp_buffer[
          0] < timestamp - self._filter_window:
        self._timestamp_buffer.pop(0)
        self._value_buffer.pop(0)
      else:
        break
    return np.mean(self._value_buffer, axis=0)



"""Load the locomotion gym env using the gin config files."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gin
from pybullet_envs.minitaur.envs_v2 import locomotion_gym_env
from pybullet_envs.minitaur.envs_v2 import multiagent_mobility_gym_env

ROBOT_FIELD_IN_CONFIG = 'robot_params'
TASK_FIELD_IN_CONFIG = 'task_params'
PMTG_FIELD_IN_CONFIG = 'pmtg_params'
_PMTG_GIN_QUERY = 'pmtg_wrapper_env.PmtgWrapperEnv.'


@gin.configurable
def load(wrapper_classes=None, multiagent=False, **kwargs):
  """load a pre-defined locomotion gym environment.

  The env specific settings should have been set in the gin files.

  Args:
    wrapper_classes: A list of wrapper_classes.
    multiagent: Whether to use multiagent environment.
    **kwargs: Keyword arguments to be passed to the environment constructor.

  Returns:
    env: The instance of the minitaur gym environment.
  """
  # Gin config are not always specified this way (e.g. namescoped config).
  # Only guery parameters when it is necessary.
  if any(
      k in (PMTG_FIELD_IN_CONFIG, TASK_FIELD_IN_CONFIG, ROBOT_FIELD_IN_CONFIG)
      for k in kwargs):
    with gin.unlock_config():
      if multiagent:
        #  Currently assume robots and tasks are identical
        robot_class = gin.query_parameter(
            'multiagent_mobility_gym_env.MultiagentMobilityGymEnv.robot_classes'
        )[0].selector
        task = gin.query_parameter(
            'multiagent_mobility_gym_env.MultiagentMobilityGymEnv.tasks'
        )[0].selector
      else:
        robot_class = gin.query_parameter(
            'locomotion_gym_env.LocomotionGymEnv.robot_class').selector
        task = gin.query_parameter(
            'locomotion_gym_env.LocomotionGymEnv.task').selector
      gin_prefix_dict = {
          PMTG_FIELD_IN_CONFIG: _PMTG_GIN_QUERY,
          TASK_FIELD_IN_CONFIG: task + '.',
          ROBOT_FIELD_IN_CONFIG: robot_class + '.',
      }
      for field_name, field_values in kwargs.items():
        if field_name in gin_prefix_dict:
          for var_name, value in field_values.items():
            gin.bind_parameter(gin_prefix_dict[field_name] + var_name, value)
        else:
          raise ValueError(
              'Environment argument type is not found in gin_prefix_dict.')
  if multiagent:
    env = multiagent_mobility_gym_env.MultiagentMobilityGymEnv()
  else:
    env = locomotion_gym_env.LocomotionGymEnv()
  if wrapper_classes is not None:
    # A little macro for the automatic list expansion
    if not isinstance(wrapper_classes, list):
      wrapper_classes = [wrapper_classes]

    # Wrap environments with user-provided wrappers
    # (e.g. TrajectoryGeneratorWrapperEnv)
    for wrapper_cls in wrapper_classes:
      env = wrapper_cls(env)
  return env



"""Utilities for safety layers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import typing
from pybullet_envs.minitaur.robots import robot_config
from pybullet_envs.minitaur.robots.safety import data_types
from pybullet_envs.minitaur.robots.safety import safety_error


def assert_in_bound(name: typing.Text, value: float, bound: data_types.Bound):
  """Check if the given value is within the provided bound.

  Args:
    name: The name of the value.
    value: Number to be checked.
    bound: Contains the lower and upper bounds. The bound is inclusive.

  Raises:
    safety_error.OutofBoundError: when the value is outside the bound.
  """
  if bound.lower <= value <= bound.upper:
    return
  else:
    raise safety_error.OutOfBoundError("{} is out of bound {} for {}".format(
        value, bound, name))


def convert_to_motor_action(
    robot: typing.Any,
    action: typing.Sequence[float],
    control_mode: robot_config.MotorControlMode,
):
  """Converts the input action to generic MotorAction classes.

  Args:
    robot: An robot instance.
    action: The motor commands sent to the robot.
    control_mode: The motor control mode.

  Returns:
    The list of converted MotorAction instances.
  """
  motor_action_list = []
  if control_mode == robot_config.MotorControlMode.POSITION:
    for motor_id, position in enumerate(action):
      motor_action_list.append(
          data_types.MotorAction(
              timestamp=robot.last_action_time,
              position=position,
              position_gain=robot.GetMotorPositionGains()[motor_id],
              velocity=0,
              velocity_gain=robot.GetMotorVelocityGains()[motor_id],
              torque=0))

  if (control_mode == robot_config.MotorControlMode.TORQUE or
      control_mode == robot_config.MotorControlMode.PWM):
    for motor_id, torque in enumerate(action):
      motor_action_list.append(
          data_types.MotorAction(
              timestamp=robot.last_action_time,
              position=0,
              position_gain=0,
              velocity=0,
              velocity_gain=0,
              torque=torque))

  if control_mode == robot_config.MotorControlMode.HYBRID:
    for motor_id in range(robot.num_motors):
      position_index = (
          motor_id * robot_config.HYBRID_ACTION_DIMENSION +
          robot_config.HybridActionIndex.POSITION.value)
      position_gain_index = (
          motor_id * robot_config.HYBRID_ACTION_DIMENSION +
          robot_config.HybridActionIndex.POSITION_GAIN.value)
      velocity_index = (
          motor_id * robot_config.HYBRID_ACTION_DIMENSION +
          robot_config.HybridActionIndex.VELOCITY.value)
      velocity_gain_index = (
          motor_id * robot_config.HYBRID_ACTION_DIMENSION +
          robot_config.HybridActionIndex.VELOCITY_GAIN.value)
      torque_index = (
          motor_id * robot_config.HYBRID_ACTION_DIMENSION +
          robot_config.HybridActionIndex.TORQUE.value)
      motor_action_list.append(
          data_types.MotorAction(
              timestamp=robot.last_action_time,
              position=action[position_index],
              position_gain=action[position_gain_index],
              velocity=action[velocity_index],
              velocity_gain=action[velocity_gain_index],
              torque=action[torque_index]))

  return motor_action_list


class MovingWindowFilter(object):
  """A stable O(1) moving filter for incoming data streams.

  We implement the Neumaier's algorithm to calculate the moving window average,
  which is numerically stable.

  """

  def __init__(self, window_size: int):
    """Initializes the class.

    Args:
      window_size: The moving window size.
    """
    assert window_size > 0
    self._window_size = window_size
    self._value_deque = collections.deque(maxlen=window_size)
    # The moving window sum.
    self._sum = 0
    # The correction term to compensate numerical precision loss during
    # calculation.
    self._correction = 0

  def _neumaier_sum(self, value: float):
    """Update the moving window sum using Neumaier's algorithm.

    For more details please refer to:
    https://en.wikipedia.org/wiki/Kahan_summation_algorithm#Further_enhancements

    Args:
      value: The new value to be added to the window.
    """

    new_sum = self._sum + value
    if abs(self._sum) >= abs(value):
      # If self._sum is bigger, low-order digits of value are lost.
      self._correction += (self._sum - new_sum) + value
    else:
      # low-order digits of sum are lost
      self._correction += (value - new_sum) + self._sum

    self._sum = new_sum

  def calculate_average(self, new_value: float) -> float:
    """Computes the moving window average in O(1) time.

    Args:
      new_value: The new value to enter the moving window.

    Returns:
      The average of the values in the window.

    """
    deque_len = len(self._value_deque)
    if deque_len < self._value_deque.maxlen:
      pass
    else:
      # The left most value to be subtracted from the moving sum.
      self._neumaier_sum(-self._value_deque[0])

    self._neumaier_sum(new_value)
    self._value_deque.append(new_value)

    return (self._sum + self._correction) / self._window_size



try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

import pybullet_envs.deep_mimic.learning.tf_util as TFUtil

NAME = "fc_2layers_1024units"


def build_net(input_tfs, reuse=False):
  layers = [1024, 512]
  activation = tf.nn.relu

  input_tf = tf.concat(axis=-1, values=input_tfs)
  h = TFUtil.fc_net(input_tf, layers, activation=activation, reuse=reuse)
  h = activation(h)
  return h



#The example to run the raibert controller in a Minitaur gym env.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(os.path.dirname(currentdir)))
os.sys.path.insert(0, parentdir)

import tf.compat.v1 as tf
from pybullet_envs.minitaur.envs import minitaur_raibert_controller
from pybullet_envs.minitaur.envs import minitaur_gym_env

flags = tf.app.flags
FLAGS = tf.app.flags.FLAGS

flags.DEFINE_float("motor_kp", 1.0, "The position gain of the motor.")
flags.DEFINE_float("motor_kd", 0.015, "The speed gain of the motor.")
flags.DEFINE_float("control_latency", 0.02, "The latency between sensor measurement and action"
                   " execution the robot.")
flags.DEFINE_string("log_path", None, "The directory to write the log file.")


def speed(t):
  max_speed = 0.35
  t1 = 3
  if t < t1:
    return t / t1 * max_speed
  else:
    return -max_speed


def main(argv):
  del argv
  try:
    env = minitaur_gym_env.MinitaurGymEnv(
        urdf_version=minitaur_gym_env.RAINBOW_DASH_V0_URDF_VERSION,
        control_time_step=0.006,
        action_repeat=6,
        pd_latency=0.003,
        control_latency=FLAGS.control_latency,
        motor_kp=FLAGS.motor_kp,
        motor_kd=FLAGS.motor_kd,
        remove_default_joint_damping=True,
        leg_model_enabled=False,
        render=True,
        on_rack=False,
        accurate_motor_model_enabled=True,
        log_path=FLAGS.log_path)
    env.reset()

    controller = minitaur_raibert_controller.MinitaurRaibertTrottingController(env.minitaur)

    tstart = env.minitaur.GetTimeSinceReset()
    for _ in range(1000):
      t = env.minitaur.GetTimeSinceReset() - tstart
      controller.behavior_parameters = (minitaur_raibert_controller.BehaviorParameters(
          desired_forward_speed=speed(t)))
      controller.update(t)
      env.step(controller.get_action())
  finally:
    env.close()


if __name__ == "__main__":
  tf.app.run(main)



from pybullet_utils import bullet_client as bc
import pybullet
import pybullet_data

p0 = bc.BulletClient(connection_mode=pybullet.DIRECT)
p0.setAdditionalSearchPath(pybullet_data.getDataPath())

p1 = bc.BulletClient(connection_mode=pybullet.DIRECT)
p1.setAdditionalSearchPath(pybullet_data.getDataPath())

#can also connect using different modes, GUI, SHARED_MEMORY, TCP, UDP, SHARED_MEMORY_SERVER, GUI_SERVER
#pgui = bc.BulletClient(connection_mode=pybullet.GUI)

p0.loadURDF("r2d2.urdf")
p1.loadSDF("stadium.sdf")
print(p0._client)
print(p1._client)
print("p0.getNumBodies()=", p0.getNumBodies())
print("p1.getNumBodies()=", p1.getNumBodies())



"""State estimator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Any, Sequence

from filterpy import kalman
import gin
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import state_estimator
from pybullet_envs.minitaur.agents.baseline_controller import time_based_moving_window_filter
from pybullet_envs.minitaur.envs_v2.sensors import accelerometer_sensor
from pybullet_envs.minitaur.envs_v2.sensors import imu_sensor

_DEFAULT_VELOCITY_FILTER_WINDOW = 0.2
_DEFAULT_GYRO_FILTER_WINDOW = 0.1
_DEFAULT_VELOCITY_CLIPPING = 0.8
_STATE_DIMENSION = 3
_GRAVITY = (0.0, 0.0, -9.8)


@gin.configurable
class IMUBasedCOMVelocityEstimator(state_estimator.StateEstimatorBase):
  """Estimate the CoM velocity using IMU sensors and velocities of stance feet.


  Estimates the com velocity of the robot using IMU data and stance feet
  velocities fused by a Kalman Filter. Kalman Filter assumes the true state x
  follows a linear dynamics: x'=Fx+Bu+w, where x' and x denots the
  current and previous state of the robot, F is the state transition matrix,
  B is the control-input matrix, and w is the noise in the dynamics, assuming to
  follow a zero-mean multivariate normal distribution with covariance Q. It also
  assumes that we can obtain a noisy observation of the state with the model
  z=Hx+v, where z is the observed state, H is the observation matrix, and v is
  a noise model, assuming to follow a zero-mean multivariate normal distribution
  with covariance R. In our case, x is the CoM velocity of the robot,
  F=B=H=eye(3), u=dt * CoM acceleration, which is obtained from accelerometer,
  and the noisy observation z is obtained from the negated average velocities at
  the end-effectors in contact with the ground.

  """

  def __init__(
      self,
      robot: Any,
      use_sensor_interface: bool = True,
      accelerometer_variance=0.1,
      observation_variance=0.1,
      initial_variance=0.1,
      velocity_filter_window: float = _DEFAULT_VELOCITY_FILTER_WINDOW,
      gyroscope_filter_window: float = _DEFAULT_GYRO_FILTER_WINDOW,
      contact_detection_threshold: float = 0.0,
      velocity_clipping: float = _DEFAULT_VELOCITY_CLIPPING,
  ):
    """Initializes the class.

    Args:
      robot: A quadruped robot.
      use_sensor_interface: Whether to use the sensor interface to obtain the
        IMU readings or directly get them from the robot class. Former should
        be used in simulation to enable added latency and noise while latter
        should be used on real robot for better performance.
      accelerometer_variance: The estimated variance in the accelerometer
        readings, used in the Kalman Filter.
      observation_variance: The estimated variance in the observed CoM velocity
        from the stance feet velocities, used in the Kalman Filter.
      initial_variance: The variance of the initial distribution for the
        estimated CoM variance.
      velocity_filter_window: The filtering window (in time) used to smooth the
        estimated CoM velocity.
      gyroscope_filter_window: The filtering window (in time) used to smooth the
        input gyroscope readings.
      contact_detection_threshold: Threshold on the contact sensor readings to
        determine whether the foot is in contact with the ground.
      velocity_clipping: Clipping value for the estimated velocity to prevent
        unrealistically large velocity estimations.
    """
    self._robot = robot
    self._contact_detection_threshold = contact_detection_threshold
    self._velocity_clipping = velocity_clipping
    self._use_sensor_interface = use_sensor_interface

    # Use the accelerometer and gyroscope sensor from the robot
    if self._use_sensor_interface:
      for sensor in self._robot.sensors:
        if isinstance(sensor, accelerometer_sensor.AccelerometerSensor):
          self._accelerometer = sensor
        if isinstance(sensor, imu_sensor.IMUSensor):
          self._gyroscope = sensor
      assert hasattr(self, "_accelerometer") and self._accelerometer is not None
      assert hasattr(self, "_gyroscope") and self._gyroscope is not None

    # x is the underlying CoM velocity we want to estimate, z is the observed
    # CoM velocity from the stance feet velocities, and u is the accelerometer
    # readings.
    self._filter = kalman.KalmanFilter(
        dim_x=_STATE_DIMENSION,
        dim_z=_STATE_DIMENSION,
        dim_u=_STATE_DIMENSION)
    # Initialize the state distribution to be a zero-mean multi-variate normal
    # distribution with initial variance.
    self._filter.x = np.zeros(_STATE_DIMENSION)
    self._initial_variance = initial_variance
    self._filter.P = np.eye(_STATE_DIMENSION) * self._initial_variance
    # Covariance matrix for the control variable.
    self._filter.Q = np.eye(_STATE_DIMENSION) * accelerometer_variance
    # Covariance matrix for the observed states.
    self._filter.R = np.eye(_STATE_DIMENSION) * observation_variance

    # observation function (z=H*x+N(0,R))
    self._filter.H = np.eye(_STATE_DIMENSION)
    # state transition matrix (x'=F*x+B*u+N(0,Q))
    self._filter.F = np.eye(_STATE_DIMENSION)
    # Control transition matrix
    self._filter.B = np.eye(_STATE_DIMENSION)

    self._velocity_filter = time_based_moving_window_filter.TimeBasedMovingWindowFilter(
        velocity_filter_window)
    self._gyroscope_filter = time_based_moving_window_filter.TimeBasedMovingWindowFilter(
        gyroscope_filter_window)

    self.reset(0)

  @property
  def com_velocity_body_yaw_aligned_frame(self) -> Sequence[float]:
    """The base velocity projected in the body yaw aligned inertial frame.

    The body yaw aligned frame is a intertia frame where the z axis coincides
    with the yaw of the robot base and the x and y axis coincides with the world
    frame. It has a zero relative velocity/angular velocity
    to the world frame.

    Returns:
      The com velocity in body yaw aligned frame.
    """
    clipped_velocity = np.clip(self._com_velocity_body_yaw_aligned_frame,
                               -self._velocity_clipping,
                               self._velocity_clipping)

    return clipped_velocity

  def reset(self, current_time):
    del current_time
    self._filter.x = np.zeros(_STATE_DIMENSION)
    self._filter.P = np.eye(_STATE_DIMENSION) * self._initial_variance

    self._com_velocity_body_yaw_aligned_frame = np.zeros(_STATE_DIMENSION)

    self._velocity_filter.reset()
    self._gyroscope_filter.reset()

    # Use None instead of 0 in case of a big gap between reset and first step.
    self._last_timestamp = None

  def update(self, current_time):
    del current_time
    current_timestamp = self._robot.timestamp

    # First time step
    if self._last_timestamp is None:
      delta_time_s = 0.0
    else:
      delta_time_s = current_timestamp - self._last_timestamp
    self._last_timestamp = current_timestamp

    if self._use_sensor_interface:
      sensor_acc = np.array(self._accelerometer.get_observation())
      gyroscope_reading = self._gyroscope.get_observation()
    else:
      sensor_acc = np.array(self._robot.base_acceleration_accelerometer)
      gyroscope_reading = self._robot.base_roll_pitch_yaw

    filtered_gyroscope_reading = self._gyroscope_filter.calculate_average(
        gyroscope_reading, current_timestamp)
    # The yaw angle is not used here because reliably estimating the yaw angle
    # of the robot is in general difficult. This leads to a body yaw aligned
    # inertia frame for the estimated velocity.
    yaw_aligned_base_orientation = self._robot.pybullet_client.getQuaternionFromEuler(
        (filtered_gyroscope_reading[0], filtered_gyroscope_reading[1], 0.0))

    rot_mat = self._robot.pybullet_client.getMatrixFromQuaternion(
        yaw_aligned_base_orientation)
    rot_mat = np.array(rot_mat).reshape((_STATE_DIMENSION, _STATE_DIMENSION))
    calibrated_acc = rot_mat.dot(sensor_acc) + np.array(_GRAVITY)
    self._filter.predict(u=calibrated_acc * delta_time_s)
    observed_velocities = []

    foot_contact = [
        np.linalg.norm(contact_force) > self._contact_detection_threshold
        for contact_force in self._robot.feet_contact_forces()
    ]
    for leg_id in range(4):
      if foot_contact[leg_id]:
        jacobian = self._robot.compute_jacobian_for_one_leg(leg_id)
        # Only pick the jacobian related to joint motors
        # TODO(magicmelon): standardize the process of picking out relevant dofs
        com_dof = self._robot.urdf_loader.com_dof
        jacobian = jacobian[:,
                            com_dof + leg_id * 3:com_dof + (leg_id + 1) * 3]
        joint_velocities = self._robot.motor_velocities[leg_id *
                                                        3:(leg_id + 1) * 3]
        leg_velocity_in_base_frame = jacobian.dot(joint_velocities)
        base_velocity_in_base_frame = -leg_velocity_in_base_frame
        observed_velocities.append(rot_mat.dot(base_velocity_in_base_frame))

    if observed_velocities:
      observed_velocities = np.mean(observed_velocities, axis=0)
      self._filter.update(observed_velocities)

    velocity = self._filter.x.copy()

    self._com_velocity_body_yaw_aligned_frame = self._velocity_filter.calculate_average(
        velocity, current_timestamp)



import time
import numpy as np
import math

useNullSpace = 1
ikSolver = 0
pandaEndEffectorIndex = 11 #8
pandaNumDofs = 7

ll = [-7]*pandaNumDofs
#upper limits for null space (todo: set them to proper range)
ul = [7]*pandaNumDofs
#joint ranges for null space (todo: set them to proper range)
jr = [7]*pandaNumDofs
#restposes for null space
jointPositions=[0.98, 0.458, 0.31, -2.24, -0.30, 2.66, 2.32, 0.02, 0.02]
rp = jointPositions

class PandaSim(object):
  def __init__(self, bullet_client, offset):
    self.bullet_client = bullet_client
    self.offset = np.array(offset)
    
    #print("offset=",offset)
    flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES
    legos=[]
    self.bullet_client.loadURDF("tray/traybox.urdf", [0+offset[0], 0+offset[1], -0.6+offset[2]], [-0.5, -0.5, -0.5, 0.5], flags=flags)
    legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.1, 0.3, -0.5])+self.offset, flags=flags))
    legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([-0.1, 0.3, -0.5])+self.offset, flags=flags))
    legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.1, 0.3, -0.7])+self.offset, flags=flags))
    sphereId = self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0, 0.3, -0.6])+self.offset, flags=flags)
    self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0, 0.3, -0.5])+self.offset, flags=flags)
    self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0, 0.3, -0.7])+self.offset, flags=flags)
    orn=[-0.707107, 0.0, 0.0, 0.707107]#p.getQuaternionFromEuler([-math.pi/2,math.pi/2,0])
    eul = self.bullet_client.getEulerFromQuaternion([-0.5, -0.5, -0.5, 0.5])
    self.panda = self.bullet_client.loadURDF("franka_panda/panda.urdf", np.array([0,0,0])+self.offset, orn, useFixedBase=True, flags=flags)
    index = 0
    for j in range(self.bullet_client.getNumJoints(self.panda)):
      self.bullet_client.changeDynamics(self.panda, j, linearDamping=0, angularDamping=0)
      info = self.bullet_client.getJointInfo(self.panda, j)
  
      jointName = info[1]
      jointType = info[2]
      if (jointType == self.bullet_client.JOINT_PRISMATIC):
        
        self.bullet_client.resetJointState(self.panda, j, jointPositions[index]) 
        index=index+1
      if (jointType == self.bullet_client.JOINT_REVOLUTE):
        self.bullet_client.resetJointState(self.panda, j, jointPositions[index]) 
        index=index+1
    self.t = 0.
  def reset(self):
    pass

  def step(self):
    t = self.t
    self.t += 1./60.
    pos = [self.offset[0]+0.2 * math.sin(1.5 * t), self.offset[1]+0.044, self.offset[2]+-0.6 + 0.1 * math.cos(1.5 * t)]
    orn = self.bullet_client.getQuaternionFromEuler([math.pi/2.,0.,0.])
    jointPoses = self.bullet_client.calculateInverseKinematics(self.panda,pandaEndEffectorIndex, pos, orn, ll, ul,
      jr, rp, maxNumIterations=5)
    for i in range(pandaNumDofs):
        self.bullet_client.setJointMotorControl2(self.panda, i, self.bullet_client.POSITION_CONTROL, jointPoses[i],force=5 * 240.)
    pass
  



"""A class for combining multiple state estimators."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Any, Sequence

import gin

from pybullet_envs.minitaur.agents.baseline_controller import state_estimator


@gin.configurable
class MultiStateEstimator(state_estimator.StateEstimatorBase):
  """Combine multiple state estimators.


  This class can be used to combine multiple state estimators into one. For
  example, one can use the COMVelocityEstimator to estimate the com velocity
  and COMHeightEstimator to estimate the com height.

  """

  def __init__(
      self,
      robot: Any,
      state_estimators: Sequence[state_estimator.StateEstimatorBase],
  ):
    self._robot = robot
    self._state_estimators = state_estimators
    self.reset(0)

  def reset(self, current_time):
    for single_state_estimator in self._state_estimators:
      single_state_estimator.reset(current_time)

  def update(self, current_time):
    for single_state_estimator in self._state_estimators:
      single_state_estimator.update(current_time)

  def __getattr__(self, attr):
    for single_state_estimator in self._state_estimators:
      if hasattr(single_state_estimator, attr):
        return getattr(single_state_estimator, attr)
    raise ValueError(
        "{} is not found in any of the state estimators".format(attr))




"""Software safety layer for robot control.

Validates the motor states received from the motor encoder.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import typing

from pybullet_envs.minitaur.robots.safety import data_types
from pybullet_envs.minitaur.robots.safety import utilities
from pybullet_envs.minitaur.robots.safety.python import moving_window_filter

# The default internal buffer size for the MotorStateValidator.
_DEQUE_SIZE = 200


class MotorStateValidator(object):
  """A safety guard to check motor states.

  Monitors the status of the motor and detects anomalies in the
  readings. For example, the class will throw safety errors if the motor
  velocity is too large. Currently we support checking of motor angle, velocity,
  gain, torque, as well as the timestamp interval.

  Attributes:
    last_state: The last received motor state.
  """

  def __init__(
      self,
      motor_id: typing.Any,
      position_bound: data_types.Bound,
      position_gain_bound: data_types.Bound,
      velocity_bound: data_types.Bound,
      velocity_gain_bound: data_types.Bound,
      torque_bound: data_types.Bound,
      timestamp_delta_bound: data_types.Bound,
      average_abs_velocity_bound: data_types.Bound,
      average_abs_power_bound: data_types.Bound,
      state_buffer_size: int = _DEQUE_SIZE,
  ):
    """Initializes the class.

    Args:
      motor_id: Unique ID for the motor.
      position_bound: The lower/upper bound of the motor angle.
      position_gain_bound: The lower/upper bound of the motor position gain for
        PD control.
      velocity_bound: The lower/upper bound of the motor speed.
      velocity_gain_bound: The lower/upper bound of the motor velocity gain for
        PD control.
      torque_bound: The lower/upper bound of the measured motor torque.
      timestamp_delta_bound: The range of timestamp difference between two
        consecutively received motor states.
      average_abs_velocity_bound: The average absolute velocity limit.
      average_abs_power_bound: The average absolute mechanical power limit.
      state_buffer_size: The buffer size used to calculate the average.
    """
    assert state_buffer_size > 1
    self.last_state = None

    self._motor_id = motor_id
    self._position_bound = position_bound
    self._position_gain_bound = position_gain_bound
    self._velocity_bound = velocity_bound
    self._velocity_gain_bound = velocity_gain_bound
    self._torque_bound = torque_bound
    self._timestamp_delta_bound = timestamp_delta_bound
    self._average_abs_velocity_bound = average_abs_velocity_bound
    self._average_abs_power_bound = average_abs_power_bound

    # For velocity/power, we use a filter to compute their averages
    # over a small period. This is to avoid the noisy readings giving false
    # positive.
    self._abs_velocity_filter = moving_window_filter.MovingWindowFilter(
        state_buffer_size)
    self._abs_power_filter = moving_window_filter.MovingWindowFilter(
        state_buffer_size)

  def on_state(self, new_state: data_types.MotorState):
    """Adds a new motor state and validates it.

    Will validate both the instantenous state as well as statitical
    averages.

    Args:
      new_state: A new state from the motor encoder.

    Raises:
      safety_error.OutOfBoundError: When any of the motor readings (e.g.
      position, torque) is out of bound.
    """

    # We first validate the new state.

    motor_str = "motor {} ".format(self._motor_id)
    utilities.assert_in_bound(motor_str + "position", new_state.position,
                              self._position_bound)
    utilities.assert_in_bound(motor_str + "velocity", new_state.velocity,
                              self._velocity_bound)
    utilities.assert_in_bound(motor_str + "position gain",
                              new_state.position_gain,
                              self._position_gain_bound)
    utilities.assert_in_bound(motor_str + "velocity gain",
                              new_state.velocity_gain,
                              self._velocity_gain_bound)
    utilities.assert_in_bound(motor_str + "torque", new_state.torque,
                              self._torque_bound)

    if not self.last_state:
      self.last_state = new_state
      return

    last_state = self.last_state

    # Check if the time interval between two received states are large.

    delta_time = new_state.timestamp - last_state.timestamp
    utilities.assert_in_bound(motor_str + "timestamp", delta_time,
                              self._timestamp_delta_bound)

    average_abs_velocity = self._abs_velocity_filter.CalculateAverage(
        abs(new_state.velocity))
    utilities.assert_in_bound(motor_str + "average velocity",
                              average_abs_velocity,
                              self._average_abs_velocity_bound)

    average_abs_power = self._abs_power_filter.CalculateAverage(
        abs(new_state.velocity * new_state.torque))
    utilities.assert_in_bound(motor_str + "average power", average_abs_power,
                              self._average_abs_power_bound)

    self.last_state = new_state



from mpi4py import MPI
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf
  
import numpy as np
import pybullet_envs.deep_mimic.learning.tf_util as TFUtil
import pybullet_utils.math_util as MathUtil
import pybullet_utils.mpi_util as MPIUtil
from pybullet_utils.logger import Logger

from pybullet_envs.deep_mimic.learning.solvers.solver import Solver


class MPISolver(Solver):
  CHECK_SYNC_ITERS = 1000

  def __init__(self, sess, optimizer, vars):
    super().__init__(vars)
    self.sess = sess
    self.optimizer = optimizer
    self._build_grad_feed(vars)
    self._update = optimizer.apply_gradients(zip(self._grad_tf_list, self.vars))
    self._set_flat_vars = TFUtil.SetFromFlat(sess, self.vars)
    self._get_flat_vars = TFUtil.GetFlat(sess, self.vars)

    self.iter = 0
    grad_dim = self._calc_grad_dim()
    self._flat_grad = np.zeros(grad_dim, dtype=np.float32)
    self._global_flat_grad = np.zeros(grad_dim, dtype=np.float32)

    return

  def get_stepsize(self):
    return self.optimizer._learning_rate_tensor.eval()

  def update(self, grads=None, grad_scale=1.0):
    if grads is not None:
      self._flat_grad = MathUtil.flatten(grads)
    else:
      self._flat_grad.fill(0)
    return self.update_flatgrad(self._flat_grad, grad_scale)

  def update_flatgrad(self, flat_grad, grad_scale=1.0):
    if self.iter % self.CHECK_SYNC_ITERS == 0:
      assert self.check_synced(), Logger.print2('Network parameters desynchronized')

    if grad_scale != 1.0:
      flat_grad *= grad_scale

    MPI.COMM_WORLD.Allreduce(flat_grad, self._global_flat_grad, op=MPI.SUM)
    self._global_flat_grad /= MPIUtil.get_num_procs()

    self._load_flat_grad(self._global_flat_grad)
    self.sess.run([self._update], self._grad_feed)
    self.iter += 1

    return

  def sync(self):
    vars = self._get_flat_vars()
    MPIUtil.bcast(vars)
    self._set_flat_vars(vars)
    return

  def check_synced(self):
    synced = True
    if self._is_root():
      vars = self._get_flat_vars()
      MPIUtil.bcast(vars)
    else:
      vars_local = self._get_flat_vars()
      vars_root = np.empty_like(vars_local)
      MPIUtil.bcast(vars_root)
      synced = (vars_local == vars_root).all()
    return synced

  def _is_root(self):
    return MPIUtil.is_root_proc()

  def _build_grad_feed(self, vars):
    self._grad_tf_list = []
    self._grad_buffers = []
    for v in self.vars:
      shape = v.get_shape()
      grad = np.zeros(shape)
      grad_tf = tf.placeholder(tf.float32, shape=shape)
      self._grad_buffers.append(grad)
      self._grad_tf_list.append(grad_tf)

    self._grad_feed = dict({g_tf: g for g_tf, g in zip(self._grad_tf_list, self._grad_buffers)})

    return

  def _calc_grad_dim(self):
    grad_dim = 0
    for grad in self._grad_buffers:
      grad_dim += grad.size
    return grad_dim

  def _load_flat_grad(self, flat_grad):
    start = 0
    for g in self._grad_buffers:
      size = g.size
      np.copyto(g, np.reshape(flat_grad[start:start + size], g.shape))
      start += size
    return



import pybullet as p
import time
import math

# This simple snake logic is from some 15 year old Havok C++ demo
# Thanks to Michael Ewert!
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
plane = p.createCollisionShape(p.GEOM_PLANE)

p.createMultiBody(0, plane)

useMaximalCoordinates = True
sphereRadius = 0.25
#colBoxId = p.createCollisionShapeArray([p.GEOM_BOX, p.GEOM_SPHERE],radii=[sphereRadius+0.03,sphereRadius+0.03], halfExtents=[[sphereRadius,sphereRadius,sphereRadius],[sphereRadius,sphereRadius,sphereRadius]])
colBoxId = p.createCollisionShape(p.GEOM_BOX,
                                  halfExtents=[sphereRadius, sphereRadius, sphereRadius])

mass = 1
visualShapeId = -1

link_Masses = []
linkCollisionShapeIndices = []
linkVisualShapeIndices = []
linkPositions = []
linkOrientations = []
linkInertialFramePositions = []
linkInertialFrameOrientations = []
indices = []
jointTypes = []
axis = []

for i in range(36):
  link_Masses.append(1)
  linkCollisionShapeIndices.append(colBoxId)
  linkVisualShapeIndices.append(-1)
  linkPositions.append([0, sphereRadius * 2.0 + 0.01, 0])
  linkOrientations.append([0, 0, 0, 1])
  linkInertialFramePositions.append([0, 0, 0])
  linkInertialFrameOrientations.append([0, 0, 0, 1])
  indices.append(i)
  jointTypes.append(p.JOINT_REVOLUTE)
  axis.append([0, 0, 1])

basePosition = [0, 0, 1]
baseOrientation = [0, 0, 0, 1]
sphereUid = p.createMultiBody(mass,
                              colBoxId,
                              visualShapeId,
                              basePosition,
                              baseOrientation,
                              linkMasses=link_Masses,
                              linkCollisionShapeIndices=linkCollisionShapeIndices,
                              linkVisualShapeIndices=linkVisualShapeIndices,
                              linkPositions=linkPositions,
                              linkOrientations=linkOrientations,
                              linkInertialFramePositions=linkInertialFramePositions,
                              linkInertialFrameOrientations=linkInertialFrameOrientations,
                              linkParentIndices=indices,
                              linkJointTypes=jointTypes,
                              linkJointAxis=axis,
                              useMaximalCoordinates=useMaximalCoordinates)

p.setGravity(0, 0, -10)
p.setRealTimeSimulation(0)

anistropicFriction = [1, 0.01, 0.01]
p.changeDynamics(sphereUid, -1, lateralFriction=2, anisotropicFriction=anistropicFriction)
p.getNumJoints(sphereUid)
for i in range(p.getNumJoints(sphereUid)):
  p.getJointInfo(sphereUid, i)
  p.changeDynamics(sphereUid, i, lateralFriction=2, anisotropicFriction=anistropicFriction)

dt = 1. / 240.
SNAKE_NORMAL_PERIOD = 0.1  #1.5
m_wavePeriod = SNAKE_NORMAL_PERIOD

m_waveLength = 4
m_wavePeriod = 1.5
m_waveAmplitude = 0.4
m_waveFront = 0.0
#our steering value
m_steering = 0.0
m_segmentLength = sphereRadius * 2.0
forward = 0

while (1):
  keys = p.getKeyboardEvents()
  for k, v in keys.items():

    if (k == p.B3G_RIGHT_ARROW and (v & p.KEY_WAS_TRIGGERED)):
      m_steering = -.2
    if (k == p.B3G_RIGHT_ARROW and (v & p.KEY_WAS_RELEASED)):
      m_steering = 0
    if (k == p.B3G_LEFT_ARROW and (v & p.KEY_WAS_TRIGGERED)):
      m_steering = .2
    if (k == p.B3G_LEFT_ARROW and (v & p.KEY_WAS_RELEASED)):
      m_steering = 0

  amp = 0.2
  offset = 0.6
  numMuscles = p.getNumJoints(sphereUid)
  scaleStart = 1.0

  #start of the snake with smaller waves.
  #I think starting the wave at the tail would work better ( while it still goes from head to tail )
  if (m_waveFront < m_segmentLength * 4.0):
    scaleStart = m_waveFront / (m_segmentLength * 4.0)

  segment = numMuscles - 1

  #we simply move a sin wave down the body of the snake.
  #this snake may be going backwards, but who can tell ;)
  for joint in range(p.getNumJoints(sphereUid)):
    segment = joint  #numMuscles-1-joint
    #map segment to phase
    phase = (m_waveFront - (segment + 1) * m_segmentLength) / m_waveLength
    phase -= math.floor(phase)
    phase *= math.pi * 2.0

    #map phase to curvature
    targetPos = math.sin(phase) * scaleStart * m_waveAmplitude

    #// steer snake by squashing +ve or -ve side of sin curve
    if (m_steering > 0 and targetPos < 0):
      targetPos *= 1.0 / (1.0 + m_steering)

    if (m_steering < 0 and targetPos > 0):
      targetPos *= 1.0 / (1.0 - m_steering)

    #set our motor
    p.setJointMotorControl2(sphereUid,
                            joint,
                            p.POSITION_CONTROL,
                            targetPosition=targetPos + m_steering,
                            force=30)

  #wave keeps track of where the wave is in time
  m_waveFront += dt / m_wavePeriod * m_waveLength
  p.stepSimulation()

  time.sleep(dt)



import time
import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)
print("parentdir=", parentdir)
import json
from pybullet_envs.deep_mimic.learning.rl_world import RLWorld
from pybullet_envs.deep_mimic.learning.ppo_agent import PPOAgent

import pybullet_data
from pybullet_utils.arg_parser import ArgParser
from pybullet_utils.logger import Logger
from pybullet_envs.deep_mimic.env.pybullet_deep_mimic_env import PyBulletDeepMimicEnv
import sys
import random

update_timestep = 1. / 240.
animating = True
step = False
total_reward = 0
steps = 0

def update_world(world, time_elapsed):
  timeStep = update_timestep
  world.update(timeStep)
  reward = world.env.calc_reward(agent_id=0)
  global total_reward
  total_reward += reward
  global steps
  steps+=1
  
  #print("reward=",reward)
  #print("steps=",steps)
  end_episode = world.env.is_episode_end()
  if (end_episode or steps>= 1000):
    print("total_reward=",total_reward)
    total_reward=0
    steps = 0
    world.end_episode()
    world.reset()
  return


def build_arg_parser(args):
  arg_parser = ArgParser()
  arg_parser.load_args(args)

  arg_file = arg_parser.parse_string('arg_file', '')
  if arg_file == '':
    arg_file = "run_humanoid3d_backflip_args.txt"
  if (arg_file != ''):
    path = pybullet_data.getDataPath() + "/args/" + arg_file
    succ = arg_parser.load_file(path)
    Logger.print2(arg_file)
    assert succ, Logger.print2('Failed to load args from: ' + arg_file)
  return arg_parser


args = sys.argv[1:]


def build_world(args, enable_draw):
  arg_parser = build_arg_parser(args)
  print("enable_draw=", enable_draw)
  env = PyBulletDeepMimicEnv(arg_parser, enable_draw)
  world = RLWorld(env, arg_parser)
  #world.env.set_playback_speed(playback_speed)

  motion_file = arg_parser.parse_string("motion_file")
  print("motion_file=", motion_file)
  bodies = arg_parser.parse_ints("fall_contact_bodies")
  print("bodies=", bodies)
  int_output_path = arg_parser.parse_string("int_output_path")
  print("int_output_path=", int_output_path)
  agent_files = pybullet_data.getDataPath() + "/" + arg_parser.parse_string("agent_files")

  AGENT_TYPE_KEY = "AgentType"

  print("agent_file=", agent_files)
  with open(agent_files) as data_file:
    json_data = json.load(data_file)
    print("json_data=", json_data)
    assert AGENT_TYPE_KEY in json_data
    agent_type = json_data[AGENT_TYPE_KEY]
    print("agent_type=", agent_type)
    agent = PPOAgent(world, id, json_data)

    agent.set_enable_training(False)
    world.reset()
  return world


if __name__ == '__main__':

  world = build_world(args, True)
  while (world.env._pybullet_client.isConnected()):

    timeStep = update_timestep
    time.sleep(timeStep)
    keys = world.env.getKeyboardEvents()

    if world.env.isKeyTriggered(keys, ' '):
      animating = not animating
    if world.env.isKeyTriggered(keys, 'i'):
      step = True
    if (animating or step):
      update_world(world, timeStep)
      step = False



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the attribute dictionary."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf

from pybullet_envs.minitaur.agents.tools import attr_dict


class AttrDictTest(tf.test.TestCase):

  def test_construct_from_dict(self):
    initial = dict(foo=13, bar=42)
    obj = attr_dict.AttrDict(initial)
    self.assertEqual(13, obj.foo)
    self.assertEqual(42, obj.bar)

  def test_construct_from_kwargs(self):
    obj = attr_dict.AttrDict(foo=13, bar=42)
    self.assertEqual(13, obj.foo)
    self.assertEqual(42, obj.bar)

  def test_has_attribute(self):
    obj = attr_dict.AttrDict(foo=13)
    self.assertTrue('foo' in obj)
    self.assertFalse('bar' in obj)

  def test_access_default(self):
    obj = attr_dict.AttrDict()
    self.assertEqual(None, obj.foo)

  def test_access_magic(self):
    obj = attr_dict.AttrDict()
    with self.assertRaises(AttributeError):
      obj.__getstate__  # pylint: disable=pointless-statement

  def test_immutable_create(self):
    obj = attr_dict.AttrDict()
    with self.assertRaises(RuntimeError):
      obj.foo = 42

  def test_immutable_modify(self):
    obj = attr_dict.AttrDict(foo=13)
    with self.assertRaises(RuntimeError):
      obj.foo = 42

  def test_immutable_unlocked(self):
    obj = attr_dict.AttrDict()
    with obj.unlocked:
      obj.foo = 42
    self.assertEqual(42, obj.foo)


if __name__ == '__main__':
  tf.test.main()



from abc import abstractmethod
import sys, abc
if sys.version_info >= (3, 4):
  ABC = abc.ABC
else:
  ABC = abc.ABCMeta('ABC', (), {})


class Solver(ABC):

  def __init__(self, vars):
    self.vars = vars
    return

  @abstractmethod
  def update(self, grads):
    pass



import numpy as np
import pybullet_envs.deep_mimic.learning.agent_builder as AgentBuilder
import pybullet_envs.deep_mimic.learning.tf_util as TFUtil
from pybullet_envs.deep_mimic.learning.rl_agent import RLAgent
from pybullet_utils.logger import Logger
import pybullet_data


class RLWorld(object):

  def __init__(self, env, arg_parser):
    TFUtil.disable_gpu()

    self.env = env
    self.arg_parser = arg_parser
    self._enable_training = True
    self.train_agents = []
    self.parse_args(arg_parser)

    self.build_agents()

    return

  def get_enable_training(self):
    return self._enable_training

  def set_enable_training(self, enable):
    self._enable_training = enable
    for i in range(len(self.agents)):
      curr_agent = self.agents[i]
      if curr_agent is not None:
        enable_curr_train = self.train_agents[i] if (len(self.train_agents) > 0) else True
        curr_agent.enable_training = self.enable_training and enable_curr_train

    if (self._enable_training):
      self.env.set_mode(RLAgent.Mode.TRAIN)
    else:
      self.env.set_mode(RLAgent.Mode.TEST)

    return

  enable_training = property(get_enable_training, set_enable_training)

  def parse_args(self, arg_parser):
    self.train_agents = self.arg_parser.parse_bools('train_agents')
    num_agents = self.env.get_num_agents()
    assert (len(self.train_agents) == num_agents or len(self.train_agents) == 0)

    return

  def shutdown(self):
    self.env.shutdown()
    return

  def build_agents(self):
    num_agents = self.env.get_num_agents()
    print("num_agents=", num_agents)
    self.agents = []

    Logger.print2('')
    Logger.print2('Num Agents: {:d}'.format(num_agents))

    agent_files = self.arg_parser.parse_strings('agent_files')
    print("len(agent_files)=", len(agent_files))
    assert (len(agent_files) == num_agents or len(agent_files) == 0)

    model_files = self.arg_parser.parse_strings('model_files')
    assert (len(model_files) == num_agents or len(model_files) == 0)

    output_path = self.arg_parser.parse_string('output_path')
    int_output_path = self.arg_parser.parse_string('int_output_path')

    for i in range(num_agents):
      curr_file = agent_files[i]
      curr_agent = self._build_agent(i, curr_file)

      if curr_agent is not None:
        curr_agent.output_dir = output_path
        curr_agent.int_output_dir = int_output_path
        Logger.print2(str(curr_agent))

        if (len(model_files) > 0):
          curr_model_file = model_files[i]
          if curr_model_file != 'none':
            curr_agent.load_model(pybullet_data.getDataPath() + "/" + curr_model_file)

      self.agents.append(curr_agent)
      Logger.print2('')

    self.set_enable_training(self.enable_training)

    return

  def update(self, timestep):
    #print("world update!\n")
    self._update_agents(timestep)
    self._update_env(timestep)
    return

  def reset(self):
    self._reset_agents()
    self._reset_env()
    return

  def end_episode(self):
    self._end_episode_agents()
    return

  def _update_env(self, timestep):
    self.env.update(timestep)
    return

  def _update_agents(self, timestep):
    #print("len(agents)=",len(self.agents))
    for agent in self.agents:
      if (agent is not None):
        agent.update(timestep)
    return

  def _reset_env(self):
    self.env.reset()
    return

  def _reset_agents(self):
    for agent in self.agents:
      if (agent != None):
        agent.reset()
    return

  def _end_episode_agents(self):
    for agent in self.agents:
      if (agent != None):
        agent.end_episode()
    return

  def _build_agent(self, id, agent_file):
    Logger.print2('Agent {:d}: {}'.format(id, agent_file))
    if (agent_file == 'none'):
      agent = None
    else:
      agent = AgentBuilder.build_agent(self, id, agent_file)
      assert (agent != None), 'Failed to build agent {:d} from: {}'.format(id, agent_file)

    return agent



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Put an OpenAI Gym environment into the TensorFlow graph."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gym
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf


class InGraphEnv(object):
  """Put an OpenAI Gym environment into the TensorFlow graph.

  The environment will be stepped and reset inside of the graph using
  tf.py_func(). The current observation, action, reward, and done flag are held
  in according variables.
  """

  def __init__(self, env):
    """Put an OpenAI Gym environment into the TensorFlow graph.

    Args:
      env: OpenAI Gym environment.
    """
    self._env = env
    observ_shape = self._parse_shape(self._env.observation_space)
    observ_dtype = self._parse_dtype(self._env.observation_space)
    action_shape = self._parse_shape(self._env.action_space)
    action_dtype = self._parse_dtype(self._env.action_space)
    with tf.name_scope('environment'):
      self._observ = tf.Variable(tf.zeros(observ_shape, observ_dtype),
                                 name='observ',
                                 trainable=False)
      self._action = tf.Variable(tf.zeros(action_shape, action_dtype),
                                 name='action',
                                 trainable=False)
      self._reward = tf.Variable(0.0, dtype=tf.float32, name='reward', trainable=False)
      self._done = tf.Variable(True, dtype=tf.bool, name='done', trainable=False)
      self._step = tf.Variable(0, dtype=tf.int32, name='step', trainable=False)

  def __getattr__(self, name):
    """Forward unimplemented attributes to the original environment.

    Args:
      name: Attribute that was accessed.

    Returns:
      Value behind the attribute name in the wrapped environment.
    """
    return getattr(self._env, name)

  def simulate(self, action):
    """Step the environment.

    The result of the step can be accessed from the variables defined below.

    Args:
      action: Tensor holding the action to apply.

    Returns:
      Operation.
    """
    with tf.name_scope('environment/simulate'):
      if action.dtype in (tf.float16, tf.float32, tf.float64):
        action = tf.check_numerics(action, 'action')
      observ_dtype = self._parse_dtype(self._env.observation_space)
      observ, reward, done = tf.py_func(lambda a: self._env.step(a)[:3], [action],
                                        [observ_dtype, tf.float32, tf.bool],
                                        name='step')
      observ = tf.check_numerics(observ, 'observ')
      reward = tf.check_numerics(reward, 'reward')
      return tf.group(self._observ.assign(observ), self._action.assign(action),
                      self._reward.assign(reward), self._done.assign(done),
                      self._step.assign_add(1))

  def reset(self):
    """Reset the environment.

    Returns:
      Tensor of the current observation.
    """
    observ_dtype = self._parse_dtype(self._env.observation_space)
    observ = tf.py_func(self._env.reset, [], observ_dtype, name='reset')
    observ = tf.check_numerics(observ, 'observ')
    with tf.control_dependencies(
        [self._observ.assign(observ),
         self._reward.assign(0),
         self._done.assign(False)]):
      return tf.identity(observ)

  @property
  def observ(self):
    """Access the variable holding the current observation."""
    return self._observ

  @property
  def action(self):
    """Access the variable holding the last recieved action."""
    return self._action

  @property
  def reward(self):
    """Access the variable holding the current reward."""
    return self._reward

  @property
  def done(self):
    """Access the variable indicating whether the episode is done."""
    return self._done

  @property
  def step(self):
    """Access the variable containg total steps of this environment."""
    return self._step

  def _parse_shape(self, space):
    """Get a tensor shape from a OpenAI Gym space.

    Args:
      space: Gym space.

    Returns:
      Shape tuple.
    """
    if isinstance(space, gym.spaces.Discrete):
      return ()
    if isinstance(space, gym.spaces.Box):
      return space.shape
    raise NotImplementedError()

  def _parse_dtype(self, space):
    """Get a tensor dtype from a OpenAI Gym space.

    Args:
      space: Gym space.

    Returns:
      TensorFlow data type.
    """
    if isinstance(space, gym.spaces.Discrete):
      return tf.int32
    if isinstance(space, gym.spaces.Box):
      return tf.float32
    raise NotImplementedError()



"""
Classic cart-pole system implemented by Rich Sutton et al.
Copied from https://webdocs.cs.ualberta.ca/~sutton/book/code/pole.c
"""
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import logging
import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import time
import subprocess
import pybullet as p2
import pybullet_data
from pybullet_utils import bullet_client as bc
from pkg_resources import parse_version
from pybullet_envs.deep_mimic.env.pybullet_deep_mimic_env import PyBulletDeepMimicEnv, InitializationStrategy
from pybullet_utils.arg_parser import ArgParser
from pybullet_utils.logger import Logger


logger = logging.getLogger(__name__)


class HumanoidDeepBulletEnv(gym.Env):
  """Base Gym environment for DeepMimic."""
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self, renders=False, arg_file='', test_mode=False,
               time_step=1./240,
               rescale_actions=True,
               rescale_observations=True):
    """    
    Args:
      test_mode (bool): in test mode, the `reset()` method will always set the mocap clip time
      to 0.
      time_step (float): physics time step.
    """
    self._arg_parser = ArgParser()
    Logger.print2("===========================================================")
    succ = False
    if (arg_file != ''):
      path = pybullet_data.getDataPath() + "/args/" + arg_file
      succ = self._arg_parser.load_file(path)
      Logger.print2(arg_file)
    assert succ, Logger.print2('Failed to load args from: ' + arg_file)

    self._p = None
    self._time_step = time_step
    self._internal_env = None
    self._renders = renders
    self._discrete_actions = False
    self._arg_file = arg_file
    self._render_height = 400
    self._render_width = 640
    self._rescale_actions = rescale_actions
    self._rescale_observations = rescale_observations
    self.agent_id = -1

    self._numSteps = None
    self.test_mode = test_mode
    if self.test_mode:
        print("Environment running in TEST mode")

    self.reset()

    # Query the policy at 30Hz
    self.policy_query_30 = True
    if self.policy_query_30:
        self._policy_step = 1./30
    else:
        self._policy_step = 1./240
    self._num_env_steps = int(self._policy_step / self._time_step)
    
    self.theta_threshold_radians = 12 * 2 * math.pi / 360
    self.x_threshold = 0.4  #2.4
    high = np.array([
        self.x_threshold * 2,
        np.finfo(np.float32).max, self.theta_threshold_radians * 2,
        np.finfo(np.float32).max
    ])

    
    ctrl_size = 43  #numDof
    root_size = 7 # root
    
    action_dim = ctrl_size - root_size
    
    action_bound_min = np.array([
        -4.79999999999, -1.00000000000, -1.00000000000, -1.00000000000, -4.00000000000,
        -1.00000000000, -1.00000000000, -1.00000000000, -7.77999999999, -1.00000000000,
        -1.000000000, -1.000000000, -7.850000000, -6.280000000, -1.000000000, -1.000000000,
        -1.000000000, -12.56000000, -1.000000000, -1.000000000, -1.000000000, -4.710000000,
        -7.779999999, -1.000000000, -1.000000000, -1.000000000, -7.850000000, -6.280000000,
        -1.000000000, -1.000000000, -1.000000000, -8.460000000, -1.000000000, -1.000000000,
        -1.000000000, -4.710000000
    ])
    
    #print("len(action_bound_min)=",len(action_bound_min))
    action_bound_max = np.array([
        4.799999999, 1.000000000, 1.000000000, 1.000000000, 4.000000000, 1.000000000, 1.000000000,
        1.000000000, 8.779999999, 1.000000000, 1.0000000, 1.0000000, 4.7100000, 6.2800000,
        1.0000000, 1.0000000, 1.0000000, 12.560000, 1.0000000, 1.0000000, 1.0000000, 7.8500000,
        8.7799999, 1.0000000, 1.0000000, 1.0000000, 4.7100000, 6.2800000, 1.0000000, 1.0000000,
        1.0000000, 10.100000, 1.0000000, 1.0000000, 1.0000000, 7.8500000
    ])
    #print("len(action_bound_max)=",len(action_bound_max))
    
    self.action_space = spaces.Box(action_bound_min, action_bound_max)
    observation_min = np.array([0.0]+[-100.0]+[-4.0]*105+[-500.0]*90)
    observation_max = np.array([1.0]+[100.0]+[4.0]*105+[500.0]*90)
    state_size = 197
    self.observation_space = spaces.Box(observation_min, observation_min, dtype=np.float32)

    self.seed()
    
    self.viewer = None
    self._configure()
    

  def _configure(self, display=None):
    self.display = display

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def step(self, action):
    agent_id = self.agent_id

    if self._rescale_actions:
      # Rescale the action
      mean = -self._action_offset
      std = 1./self._action_scale
      action = action * std + mean

    # Record reward
    reward = self._internal_env.calc_reward(agent_id)

    # Apply control action
    self._internal_env.set_action(agent_id, action)

    start_time = self._internal_env.t

    # step sim
    for i in range(self._num_env_steps):
        self._internal_env.update(self._time_step)

    elapsed_time = self._internal_env.t - start_time

    self._numSteps += 1

    # Record state
    self.state = self._internal_env.record_state(agent_id)
    
    if self._rescale_observations:
      state = np.array(self.state)
      mean = -self._state_offset
      std = 1./self._state_scale
      state = (state - mean) / (std + 1e-8)

    # Record done
    done = self._internal_env.is_episode_end()
    
    info = {}
    return state, reward, done, info

  def reset(self):
    # use the initialization strategy
    if self._internal_env is None:
      if self.test_mode:
        init_strat = InitializationStrategy.START
      else:
        init_strat = InitializationStrategy.RANDOM
      self._internal_env = PyBulletDeepMimicEnv(self._arg_parser, self._renders,
                                                time_step=self._time_step,
                                                init_strategy=init_strat)

    self._internal_env.reset()
    self._p = self._internal_env._pybullet_client
    agent_id = self.agent_id  # unused here
    self._state_offset = self._internal_env.build_state_offset(self.agent_id)
    self._state_scale = self._internal_env.build_state_scale(self.agent_id)
    self._action_offset = self._internal_env.build_action_offset(self.agent_id)
    self._action_scale = self._internal_env.build_action_scale(self.agent_id)
    self._numSteps = 0
    # Record state
    self.state = self._internal_env.record_state(agent_id)

    # return state as ndarray
    state = np.array(self.state)
    if self._rescale_observations:
      mean = -self._state_offset
      std = 1./self._state_scale
      state = (state - mean) / (std + 1e-8)
    return state

  def render(self, mode='human', close=False):
    if mode == "human":
      self._renders = True
    if mode != "rgb_array":
      return np.array([])
    human = self._internal_env._humanoid
    base_pos, orn = self._p.getBasePositionAndOrientation(human._sim_model)
    base_pos = np.asarray(base_pos)
    # track the position
    base_pos[1] += 0.3
    rpy = self._p.getEulerFromQuaternion(orn)  # rpy, in radians
    rpy = 180 / np.pi * np.asarray(rpy)  # convert rpy in degrees

    self._cam_dist = 3
    self._cam_pitch = 0.3
    self._cam_yaw = 0
    if (not self._p == None):
      view_matrix = self._p.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=base_pos,
        distance=self._cam_dist,
        yaw=self._cam_yaw,
        pitch=self._cam_pitch,
        roll=0,
        upAxisIndex=1)
      proj_matrix = self._p.computeProjectionMatrixFOV(fov=60,
             aspect=float(self._render_width) / self._render_height,
             nearVal=0.1,
             farVal=100.0)
      (_, _, px, _, _) = self._p.getCameraImage(
          width=self._render_width,
          height=self._render_height,
          renderer=self._p.ER_BULLET_HARDWARE_OPENGL,
          viewMatrix=view_matrix,
          projectionMatrix=proj_matrix)
      # self._p.resetDebugVisualizerCamera(
      #   cameraDistance=2 * self._cam_dist,
      #   cameraYaw=self._cam_yaw,
      #   cameraPitch=self._cam_pitch,
      #   cameraTargetPosition=base_pos
      # )
    else:
      px = np.array([[[255,255,255,255]]*self._render_width]*self._render_height, dtype=np.uint8)
    rgb_array = np.array(px, dtype=np.uint8)
    rgb_array = np.reshape(np.array(px), (self._render_height, self._render_width, -1))
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def configure(self, args):
    pass
    
  def close(self):
    
    pass

class HumanoidDeepMimicBackflipBulletEnv(HumanoidDeepBulletEnv):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self, renders=False):
    # start the bullet physics server
    HumanoidDeepBulletEnv.__init__(self, renders, arg_file="run_humanoid3d_backflip_args.txt")


class HumanoidDeepMimicWalkBulletEnv(HumanoidDeepBulletEnv):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self, renders=False):
    # start the bullet physics server
    HumanoidDeepBulletEnv.__init__(self, renders, arg_file="run_humanoid3d_walk_args.txt")

class CartPoleContinuousBulletEnv5(HumanoidDeepBulletEnv):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self, renders=False):
    # start the bullet physics server
    HumanoidDeepBulletEnv.__init__(self, renders, arg_file="")



"""An environment randomizer that randomizes physical parameters from config."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import random

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
parentdir = os.path.dirname(os.path.dirname(parentdir))
os.sys.path.insert(0, parentdir)

import numpy as np
import tf.compat.v1 as tf
from pybullet_envs.minitaur.envs import env_randomizer_base
from pybullet_envs.minitaur.envs.env_randomizers import minitaur_env_randomizer_config

SIMULATION_TIME_STEP = 0.001


class MinitaurEnvRandomizerFromConfig(env_randomizer_base.EnvRandomizerBase):
  """A randomizer that change the minitaur_gym_env during every reset."""

  def __init__(self, config=None):
    if config is None:
      config = "all_params"
    try:
      config = getattr(minitaur_env_randomizer_config, config)
    except AttributeError:
      raise ValueError("Config {} is not found.".format(config))
    self._randomization_param_dict = config()
    tf.logging.info("Randomization config is: {}".format(self._randomization_param_dict))

  def randomize_env(self, env):
    """Randomize various physical properties of the environment.

    It randomizes the physical parameters according to the input configuration.

    Args:
      env: A minitaur gym environment.
    """
    self._randomization_function_dict = self._build_randomization_function_dict(env)
    for param_name, random_range in self._randomization_param_dict.iteritems():
      self._randomization_function_dict[param_name](lower_bound=random_range[0],
                                                    upper_bound=random_range[1])

  def _build_randomization_function_dict(self, env):
    func_dict = {}
    func_dict["mass"] = functools.partial(self._randomize_masses, minitaur=env.minitaur)
    func_dict["inertia"] = functools.partial(self._randomize_inertia, minitaur=env.minitaur)
    func_dict["latency"] = functools.partial(self._randomize_latency, minitaur=env.minitaur)
    func_dict["joint friction"] = functools.partial(self._randomize_joint_friction,
                                                    minitaur=env.minitaur)
    func_dict["motor friction"] = functools.partial(self._randomize_motor_friction,
                                                    minitaur=env.minitaur)
    func_dict["restitution"] = functools.partial(self._randomize_contact_restitution,
                                                 minitaur=env.minitaur)
    func_dict["lateral friction"] = functools.partial(self._randomize_contact_friction,
                                                      minitaur=env.minitaur)
    func_dict["battery"] = functools.partial(self._randomize_battery_level, minitaur=env.minitaur)
    func_dict["motor strength"] = functools.partial(self._randomize_motor_strength,
                                                    minitaur=env.minitaur)
    # Settinmg control step needs access to the environment.
    func_dict["control step"] = functools.partial(self._randomize_control_step, env=env)
    return func_dict

  def _randomize_control_step(self, env, lower_bound, upper_bound):
    randomized_control_step = random.uniform(lower_bound, upper_bound)
    env.set_time_step(randomized_control_step)
    tf.logging.info("control step is: {}".format(randomized_control_step))

  def _randomize_masses(self, minitaur, lower_bound, upper_bound):
    base_mass = minitaur.GetBaseMassesFromURDF()
    random_base_ratio = random.uniform(lower_bound, upper_bound)
    randomized_base_mass = random_base_ratio * np.array(base_mass)
    minitaur.SetBaseMasses(randomized_base_mass)
    tf.logging.info("base mass is: {}".format(randomized_base_mass))

    leg_masses = minitaur.GetLegMassesFromURDF()
    random_leg_ratio = random.uniform(lower_bound, upper_bound)
    randomized_leg_masses = random_leg_ratio * np.array(leg_masses)
    minitaur.SetLegMasses(randomized_leg_masses)
    tf.logging.info("leg mass is: {}".format(randomized_leg_masses))

  def _randomize_inertia(self, minitaur, lower_bound, upper_bound):
    base_inertia = minitaur.GetBaseInertiasFromURDF()
    random_base_ratio = random.uniform(lower_bound, upper_bound)
    randomized_base_inertia = random_base_ratio * np.array(base_inertia)
    minitaur.SetBaseInertias(randomized_base_inertia)
    tf.logging.info("base inertia is: {}".format(randomized_base_inertia))
    leg_inertia = minitaur.GetLegInertiasFromURDF()
    random_leg_ratio = random.uniform(lower_bound, upper_bound)
    randomized_leg_inertia = random_leg_ratio * np.array(leg_inertia)
    minitaur.SetLegInertias(randomized_leg_inertia)
    tf.logging.info("leg inertia is: {}".format(randomized_leg_inertia))

  def _randomize_latency(self, minitaur, lower_bound, upper_bound):
    randomized_latency = random.uniform(lower_bound, upper_bound)
    minitaur.SetControlLatency(randomized_latency)
    tf.logging.info("control latency is: {}".format(randomized_latency))

  def _randomize_joint_friction(self, minitaur, lower_bound, upper_bound):
    num_knee_joints = minitaur.GetNumKneeJoints()
    randomized_joint_frictions = np.random.uniform([lower_bound] * num_knee_joints,
                                                   [upper_bound] * num_knee_joints)
    minitaur.SetJointFriction(randomized_joint_frictions)
    tf.logging.info("joint friction is: {}".format(randomized_joint_frictions))

  def _randomize_motor_friction(self, minitaur, lower_bound, upper_bound):
    randomized_motor_damping = random.uniform(lower_bound, upper_bound)
    minitaur.SetMotorViscousDamping(randomized_motor_damping)
    tf.logging.info("motor friction is: {}".format(randomized_motor_damping))

  def _randomize_contact_restitution(self, minitaur, lower_bound, upper_bound):
    randomized_restitution = random.uniform(lower_bound, upper_bound)
    minitaur.SetFootRestitution(randomized_restitution)
    tf.logging.info("foot restitution is: {}".format(randomized_restitution))

  def _randomize_contact_friction(self, minitaur, lower_bound, upper_bound):
    randomized_foot_friction = random.uniform(lower_bound, upper_bound)
    minitaur.SetFootFriction(randomized_foot_friction)
    tf.logging.info("foot friction is: {}".format(randomized_foot_friction))

  def _randomize_battery_level(self, minitaur, lower_bound, upper_bound):
    randomized_battery_voltage = random.uniform(lower_bound, upper_bound)
    minitaur.SetBatteryVoltage(randomized_battery_voltage)
    tf.logging.info("battery voltage is: {}".format(randomized_battery_voltage))

  def _randomize_motor_strength(self, minitaur, lower_bound, upper_bound):
    randomized_motor_strength_ratios = np.random.uniform([lower_bound] * minitaur.num_motors,
                                                         [upper_bound] * minitaur.num_motors)
    minitaur.SetMotorStrengthRatios(randomized_motor_strength_ratios)
    tf.logging.info("motor strength is: {}".format(randomized_motor_strength_ratios))



"""Pybullet simulation of a Laikago robot."""
import math
import os
import re
import gin
import numpy as np
from pybullet_utils import transformations
from pybullet_envs.minitaur.envs_v2.utilities import laikago_pose_utils
from pybullet_envs.minitaur.robots import laikago_constants
from pybullet_envs.minitaur.robots import laikago_motor
from pybullet_envs.minitaur.robots import minitaur
from pybullet_envs.minitaur.robots import robot_config

NUM_MOTORS = 12
NUM_LEGS = 4
MOTOR_NAMES = [
    "FR_hip_motor_2_chassis_joint",
    "FR_upper_leg_2_hip_motor_joint",
    "FR_lower_leg_2_upper_leg_joint",
    "FL_hip_motor_2_chassis_joint",
    "FL_upper_leg_2_hip_motor_joint",
    "FL_lower_leg_2_upper_leg_joint",
    "RR_hip_motor_2_chassis_joint",
    "RR_upper_leg_2_hip_motor_joint",
    "RR_lower_leg_2_upper_leg_joint",
    "RL_hip_motor_2_chassis_joint",
    "RL_upper_leg_2_hip_motor_joint",
    "RL_lower_leg_2_upper_leg_joint",
]
INIT_RACK_POSITION = [0, 0, 1]
INIT_POSITION = [0, 0, 0.48]
JOINT_DIRECTIONS = np.array([-1, 1, 1, 1, 1, 1, -1, 1, 1, 1, 1, 1])
HIP_JOINT_OFFSET = 0.0
UPPER_LEG_JOINT_OFFSET = -0.6
KNEE_JOINT_OFFSET = 0.66
DOFS_PER_LEG = 3
JOINT_OFFSETS = np.array(
    [HIP_JOINT_OFFSET, UPPER_LEG_JOINT_OFFSET, KNEE_JOINT_OFFSET] * 4)
PI = math.pi

MAX_MOTOR_ANGLE_CHANGE_PER_STEP = 0.12
_DEFAULT_HIP_POSITIONS = (
    (0.21, -0.1157, 0),
    (0.21, 0.1157, 0),
    (-0.21, -0.1157, 0),
    (-0.21, 0.1157, 0),
)

# Bases on the readings from Laikago's default pose.
INIT_MOTOR_ANGLES = np.array([
    laikago_pose_utils.LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
    laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
    laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE
] * NUM_LEGS)

CHASSIS_NAME_PATTERN = re.compile(r"\w+_chassis_\w+")
MOTOR_NAME_PATTERN = re.compile(r"\w+_hip_motor_\w+")
KNEE_NAME_PATTERN = re.compile(r"\w+_lower_leg_\w+")
TOE_NAME_PATTERN = re.compile(r"jtoe\d*")

URDF_NO_TOES = "laikago.urdf"
URDF_WITH_TOES = "laikago_toes_zup.urdf"

_BODY_B_FIELD_NUMBER = 2
_LINK_A_FIELD_NUMBER = 3


@gin.configurable
class Laikago(minitaur.Minitaur):
  """A simulation for the Laikago robot."""

  def __init__(self, urdf_filename=URDF_WITH_TOES, **kwargs):
    self._urdf_filename = urdf_filename
    if "motor_kp" not in kwargs:
      kwargs["motor_kp"] = 100.0
    if "motor_kd" not in kwargs:
      kwargs["motor_kd"] = 2.0
    if "motor_torque_limits" not in kwargs:
      kwargs["motor_torque_limits"] = None

    # enable_clip_motor_commands: Boolean indicating if clipping should be
    # applied to motor commands, which limits the amount of change in joint
    # pose between timesteps.
    if "enable_clip_motor_commands" in kwargs:
      self._enable_clip_motor_commands = kwargs["enable_clip_motor_commands"]
      del kwargs["enable_clip_motor_commands"]
    else:
      self._enable_clip_motor_commands = False

    # The follwing parameters are fixed for the Laikago robot.
    kwargs["num_motors"] = NUM_MOTORS
    kwargs["dofs_per_leg"] = DOFS_PER_LEG
    kwargs["motor_direction"] = JOINT_DIRECTIONS
    kwargs["motor_offset"] = JOINT_OFFSETS
    kwargs["motor_overheat_protection"] = False
    kwargs["motor_model_class"] = laikago_motor.LaikagoMotorModel
    kwargs["safety_config"] = None

    super(Laikago, self).__init__(**kwargs)

  def _LoadRobotURDF(self):
    laikago_urdf_path = self.GetURDFFile()
    if self._self_collision_enabled:
      self.quadruped = self._pybullet_client.loadURDF(
          laikago_urdf_path,
          self._GetDefaultInitPosition(),
          self._GetDefaultInitOrientation(),
          flags=self._pybullet_client.URDF_USE_SELF_COLLISION)
    else:
      self.quadruped = self._pybullet_client.loadURDF(
          laikago_urdf_path, self._GetDefaultInitPosition(),
          self._GetDefaultInitOrientation())

  def _SettleDownForReset(self, default_motor_angles, reset_time):
    self.ReceiveObservation()

    if reset_time <= 0:
      return

    for _ in range(500):
      self._StepInternal(
          INIT_MOTOR_ANGLES,
          motor_control_mode=robot_config.MotorControlMode.POSITION)
    if default_motor_angles is not None:
      num_steps_to_reset = int(reset_time / self.time_step)
      for _ in range(num_steps_to_reset):
        self._StepInternal(
            default_motor_angles,
            motor_control_mode=robot_config.MotorControlMode.POSITION)

  def GetHipPositionsInBaseFrame(self):
    return _DEFAULT_HIP_POSITIONS

  def GetFootContacts(self):
    all_contacts = self._pybullet_client.getContactPoints(bodyA=self.quadruped)

    contacts = [False, False, False, False]
    for contact in all_contacts:
      # Ignore self contacts
      if contact[_BODY_B_FIELD_NUMBER] == self.quadruped:
        continue
      try:
        toe_link_index = self._foot_link_ids.index(
            contact[_LINK_A_FIELD_NUMBER])
        contacts[toe_link_index] = True
      except ValueError:
        continue
    return contacts

  def ComputeJacobian(self, leg_id):
    """Compute the Jacobian for a given leg."""
    # Because of the default rotation in the Laikago URDF, we need to reorder
    # the rows in the Jacobian matrix.
    if self._urdf_filename == URDF_WITH_TOES:
      return super(Laikago, self).ComputeJacobian(leg_id)
    else:
      return super(Laikago, self).ComputeJacobian(leg_id)[(2, 0, 1), :]

  def ResetPose(self, add_constraint):
    del add_constraint
    for name in self._joint_name_to_id:
      joint_id = self._joint_name_to_id[name]
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=(joint_id),
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=0)
    for name, i in zip(MOTOR_NAMES, range(len(MOTOR_NAMES))):
      if "hip_motor_2_chassis_joint" in name:
        angle = INIT_MOTOR_ANGLES[i] + HIP_JOINT_OFFSET
      elif "upper_leg_2_hip_motor_joint" in name:
        angle = INIT_MOTOR_ANGLES[i] + UPPER_LEG_JOINT_OFFSET
      elif "lower_leg_2_upper_leg_joint" in name:
        angle = INIT_MOTOR_ANGLES[i] + KNEE_JOINT_OFFSET
      else:
        raise ValueError("The name %s is not recognized as a motor joint." %
                         name)
      self._pybullet_client.resetJointState(
          self.quadruped, self._joint_name_to_id[name], angle, targetVelocity=0)

  def GetURDFFile(self):
    return os.path.join(self._urdf_root, "laikago/" + self._urdf_filename)

  def _BuildUrdfIds(self):
    """Build the link Ids from its name in the URDF file.

    Raises:
      ValueError: Unknown category of the joint name.
    """
    num_joints = self._pybullet_client.getNumJoints(self.quadruped)
    self._chassis_link_ids = [-1]
    self._leg_link_ids = []
    self._motor_link_ids = []
    self._knee_link_ids = []
    self._foot_link_ids = []

    for i in range(num_joints):
      joint_info = self._pybullet_client.getJointInfo(self.quadruped, i)
      joint_name = joint_info[1].decode("UTF-8")
      joint_id = self._joint_name_to_id[joint_name]
      if CHASSIS_NAME_PATTERN.match(joint_name):
        self._chassis_link_ids.append(joint_id)
      elif MOTOR_NAME_PATTERN.match(joint_name):
        self._motor_link_ids.append(joint_id)
      # We either treat the lower leg or the toe as the foot link, depending on
      # the urdf version used.
      elif KNEE_NAME_PATTERN.match(joint_name):
        self._knee_link_ids.append(joint_id)
      elif TOE_NAME_PATTERN.match(joint_name):
        assert self._urdf_filename == URDF_WITH_TOES
        self._foot_link_ids.append(joint_id)
      else:
        raise ValueError("Unknown category of joint %s" % joint_name)

    self._leg_link_ids.extend(self._knee_link_ids)
    self._leg_link_ids.extend(self._foot_link_ids)

    if self._urdf_filename == URDF_NO_TOES:
      self._foot_link_ids.extend(self._knee_link_ids)

    assert len(self._foot_link_ids) == NUM_LEGS
    self._chassis_link_ids.sort()
    self._motor_link_ids.sort()
    self._knee_link_ids.sort()
    self._foot_link_ids.sort()
    self._leg_link_ids.sort()

    return

  def _GetMotorNames(self):
    return MOTOR_NAMES

  def _GetDefaultInitPosition(self):
    if self._on_rack:
      return INIT_RACK_POSITION
    else:
      return INIT_POSITION

  def _GetDefaultInitOrientation(self):
    # The Laikago URDF assumes the initial pose of heading towards z axis,
    # and belly towards y axis. The following transformation is to transform
    # the Laikago initial orientation to our commonly used orientation: heading
    # towards -x direction, and z axis is the up direction.
    if self._urdf_filename == URDF_WITH_TOES:
      return [0, 0, 0, 1]
    else:
      return transformations.quaternion_from_euler(
          ai=math.pi / 2.0, aj=0, ak=math.pi / 2.0, axes="sxyz")

  def GetDefaultInitPosition(self):
    """Get default initial base position."""
    return self._GetDefaultInitPosition()

  def GetDefaultInitOrientation(self):
    """Get default initial base orientation."""
    return self._GetDefaultInitOrientation()

  def GetDefaultInitJointPose(self):
    """Get default initial joint pose."""
    joint_pose = (INIT_MOTOR_ANGLES + JOINT_OFFSETS) * JOINT_DIRECTIONS
    return joint_pose

  def ApplyAction(self, motor_commands, motor_control_mode=None):
    """Clips and then apply the motor commands using the motor model.

    Args:
      motor_commands: np.array. Can be motor angles, torques, hybrid commands,
        or motor pwms (for Minitaur only).N
      motor_control_mode: A MotorControlMode enum.
    """
    if self._enable_clip_motor_commands:
      motor_commands = self._ClipMotorCommands(motor_commands)

    super(Laikago, self).ApplyAction(motor_commands, motor_control_mode)
    return

  def _ClipMotorCommands(self, motor_commands):
    """Clips motor commands.

    Args:
      motor_commands: np.array. Can be motor angles, torques, hybrid commands,
        or motor pwms (for Minitaur only).

    Returns:
      Clipped motor commands.
    """

    # clamp the motor command by the joint limit, in case weired things happens
    max_angle_change = MAX_MOTOR_ANGLE_CHANGE_PER_STEP
    current_motor_angles = self.GetMotorAngles()
    motor_commands = np.clip(motor_commands,
                             current_motor_angles - max_angle_change,
                             current_motor_angles + max_angle_change)
    return motor_commands

  @classmethod
  def GetConstants(cls):
    del cls
    return laikago_constants

  # The following functions are added for the migration purpose. Will be removed
  # after the migration is complete.

  @property
  def robot_id(self):
    return self.quadruped

  @property
  def base_position(self):
    return self.GetBasePosition()

  @property
  def base_roll_pitch_yaw(self):
    return self.GetTrueBaseRollPitchYaw()

  @property
  def timestamp(self):
    return self.GetTimeSinceReset()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vector.proto

import sys
_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='vector.proto',
    package='robotics.messages',
    syntax='proto3',
    serialized_pb=_b(
        '\n\x0cvector.proto\x12\x11robotics.messages\"6\n\x08Vector4d\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\x12\t\n\x01w\x18\x04 \x01(\x01\"6\n\x08Vector4f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\x12\t\n\x01w\x18\x04 \x01(\x02\"+\n\x08Vector3d\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\"+\n\x08Vector3f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\" \n\x08Vector2d\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\" \n\x08Vector2f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\"\x1b\n\x07Vectord\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x01\x42\x02\x10\x01\"\x1b\n\x07Vectorf\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x02\x42\x02\x10\x01\x62\x06proto3'
    ))

_VECTOR4D = _descriptor.Descriptor(
    name='Vector4d',
    full_name='robotics.messages.Vector4d',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='x',
                                    full_name='robotics.messages.Vector4d.x',
                                    index=0,
                                    number=1,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='y',
                                    full_name='robotics.messages.Vector4d.y',
                                    index=1,
                                    number=2,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='z',
                                    full_name='robotics.messages.Vector4d.z',
                                    index=2,
                                    number=3,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='w',
                                    full_name='robotics.messages.Vector4d.w',
                                    index=3,
                                    number=4,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=35,
    serialized_end=89,
)

_VECTOR4F = _descriptor.Descriptor(
    name='Vector4f',
    full_name='robotics.messages.Vector4f',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='x',
                                    full_name='robotics.messages.Vector4f.x',
                                    index=0,
                                    number=1,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='y',
                                    full_name='robotics.messages.Vector4f.y',
                                    index=1,
                                    number=2,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='z',
                                    full_name='robotics.messages.Vector4f.z',
                                    index=2,
                                    number=3,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='w',
                                    full_name='robotics.messages.Vector4f.w',
                                    index=3,
                                    number=4,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=91,
    serialized_end=145,
)

_VECTOR3D = _descriptor.Descriptor(
    name='Vector3d',
    full_name='robotics.messages.Vector3d',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='x',
                                    full_name='robotics.messages.Vector3d.x',
                                    index=0,
                                    number=1,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='y',
                                    full_name='robotics.messages.Vector3d.y',
                                    index=1,
                                    number=2,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='z',
                                    full_name='robotics.messages.Vector3d.z',
                                    index=2,
                                    number=3,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=147,
    serialized_end=190,
)

_VECTOR3F = _descriptor.Descriptor(
    name='Vector3f',
    full_name='robotics.messages.Vector3f',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='x',
                                    full_name='robotics.messages.Vector3f.x',
                                    index=0,
                                    number=1,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='y',
                                    full_name='robotics.messages.Vector3f.y',
                                    index=1,
                                    number=2,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='z',
                                    full_name='robotics.messages.Vector3f.z',
                                    index=2,
                                    number=3,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=192,
    serialized_end=235,
)

_VECTOR2D = _descriptor.Descriptor(
    name='Vector2d',
    full_name='robotics.messages.Vector2d',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='x',
                                    full_name='robotics.messages.Vector2d.x',
                                    index=0,
                                    number=1,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='y',
                                    full_name='robotics.messages.Vector2d.y',
                                    index=1,
                                    number=2,
                                    type=1,
                                    cpp_type=5,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=237,
    serialized_end=269,
)

_VECTOR2F = _descriptor.Descriptor(
    name='Vector2f',
    full_name='robotics.messages.Vector2f',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='x',
                                    full_name='robotics.messages.Vector2f.x',
                                    index=0,
                                    number=1,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
        _descriptor.FieldDescriptor(name='y',
                                    full_name='robotics.messages.Vector2f.y',
                                    index=1,
                                    number=2,
                                    type=2,
                                    cpp_type=6,
                                    label=1,
                                    has_default_value=False,
                                    default_value=float(0),
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=None,
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=271,
    serialized_end=303,
)

_VECTORD = _descriptor.Descriptor(
    name='Vectord',
    full_name='robotics.messages.Vectord',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='data',
                                    full_name='robotics.messages.Vectord.data',
                                    index=0,
                                    number=1,
                                    type=1,
                                    cpp_type=5,
                                    label=3,
                                    has_default_value=False,
                                    default_value=[],
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=_descriptor._ParseOptions(
                                        descriptor_pb2.FieldOptions(), _b('\020\001')),
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=305,
    serialized_end=332,
)

_VECTORF = _descriptor.Descriptor(
    name='Vectorf',
    full_name='robotics.messages.Vectorf',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(name='data',
                                    full_name='robotics.messages.Vectorf.data',
                                    index=0,
                                    number=1,
                                    type=2,
                                    cpp_type=6,
                                    label=3,
                                    has_default_value=False,
                                    default_value=[],
                                    message_type=None,
                                    enum_type=None,
                                    containing_type=None,
                                    is_extension=False,
                                    extension_scope=None,
                                    options=_descriptor._ParseOptions(
                                        descriptor_pb2.FieldOptions(), _b('\020\001')),
                                    file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=334,
    serialized_end=361,
)

DESCRIPTOR.message_types_by_name['Vector4d'] = _VECTOR4D
DESCRIPTOR.message_types_by_name['Vector4f'] = _VECTOR4F
DESCRIPTOR.message_types_by_name['Vector3d'] = _VECTOR3D
DESCRIPTOR.message_types_by_name['Vector3f'] = _VECTOR3F
DESCRIPTOR.message_types_by_name['Vector2d'] = _VECTOR2D
DESCRIPTOR.message_types_by_name['Vector2f'] = _VECTOR2F
DESCRIPTOR.message_types_by_name['Vectord'] = _VECTORD
DESCRIPTOR.message_types_by_name['Vectorf'] = _VECTORF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Vector4d = _reflection.GeneratedProtocolMessageType(
    'Vector4d',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTOR4D,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vector4d)
        ))
_sym_db.RegisterMessage(Vector4d)

Vector4f = _reflection.GeneratedProtocolMessageType(
    'Vector4f',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTOR4F,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vector4f)
        ))
_sym_db.RegisterMessage(Vector4f)

Vector3d = _reflection.GeneratedProtocolMessageType(
    'Vector3d',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTOR3D,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vector3d)
        ))
_sym_db.RegisterMessage(Vector3d)

Vector3f = _reflection.GeneratedProtocolMessageType(
    'Vector3f',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTOR3F,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vector3f)
        ))
_sym_db.RegisterMessage(Vector3f)

Vector2d = _reflection.GeneratedProtocolMessageType(
    'Vector2d',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTOR2D,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vector2d)
        ))
_sym_db.RegisterMessage(Vector2d)

Vector2f = _reflection.GeneratedProtocolMessageType(
    'Vector2f',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTOR2F,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vector2f)
        ))
_sym_db.RegisterMessage(Vector2f)

Vectord = _reflection.GeneratedProtocolMessageType(
    'Vectord',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTORD,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vectord)
        ))
_sym_db.RegisterMessage(Vectord)

Vectorf = _reflection.GeneratedProtocolMessageType(
    'Vectorf',
    (_message.Message,),
    dict(DESCRIPTOR=_VECTORF,
         __module__='vector_pb2'
         # @@protoc_insertion_point(class_scope:robotics.messages.Vectorf)
        ))
_sym_db.RegisterMessage(Vectorf)

_VECTORD.fields_by_name['data'].has_options = True
_VECTORD.fields_by_name['data']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(),
                                                                     _b('\020\001'))
_VECTORF.fields_by_name['data'].has_options = True
_VECTORF.fields_by_name['data']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(),
                                                                     _b('\020\001'))
# @@protoc_insertion_point(module_scope)



import pybullet as p
import pybullet_data as pd
import math
import time
import numpy as np
import pybullet_robots.panda.panda_sim_grasp as panda_sim

#video requires ffmpeg available in path
createVideo=False
fps=240.
timeStep = 1./fps

if createVideo:
	p.connect(p.GUI, options="--minGraphicsUpdateTimeMs=0 --mp4=\"pybullet_grasp.mp4\" --mp4fps="+str(fps) )
else:
	p.connect(p.GUI)

p.configureDebugVisualizer(p.COV_ENABLE_Y_AXIS_UP,1)
p.configureDebugVisualizer(p.COV_ENABLE_GUI,0)
p.setPhysicsEngineParameter(maxNumCmdPer1ms=1000)
p.resetDebugVisualizerCamera(cameraDistance=1.3, cameraYaw=38, cameraPitch=-22, cameraTargetPosition=[0.35,-0.13,0])
p.setAdditionalSearchPath(pd.getDataPath())

p.setTimeStep(timeStep)
p.setGravity(0,-9.8,0)

panda = panda_sim.PandaSimAuto(p,[0,0,0])
panda.control_dt = timeStep

logId = panda.bullet_client.startStateLogging(panda.bullet_client.STATE_LOGGING_PROFILE_TIMINGS, "log.json")
panda.bullet_client.submitProfileTiming("start")
for i in range (100000):
	panda.bullet_client.submitProfileTiming("full_step")
	panda.step()
	p.stepSimulation()
	if createVideo:
		p.configureDebugVisualizer(p.COV_ENABLE_SINGLE_STEP_RENDERING,1)
	if not createVideo:
		time.sleep(timeStep)
	panda.bullet_client.submitProfileTiming()
panda.bullet_client.submitProfileTiming()
panda.bullet_client.stopStateLogging(logId)
	



import numpy as np
import sys
import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)
print("parentdir=", parentdir)

from pybullet_envs.deep_mimic.env.pybullet_deep_mimic_env import PyBulletDeepMimicEnv
from pybullet_envs.deep_mimic.learning.rl_world import RLWorld
from pybullet_utils.logger import Logger
from pybullet_envs.deep_mimic.testrl import update_world, update_timestep, build_world
import pybullet_utils.mpi_util as MPIUtil

args = []
world = None


def run():
  global update_timestep
  global world
  done = False
  while not (done):
    update_world(world, update_timestep)

  return


def shutdown():
  global world

  Logger.print2('Shutting down...')
  world.shutdown()
  return


def main():
  global args
  global world

  # Command line arguments
  args = sys.argv[1:]
  enable_draw = False
  world = build_world(args, enable_draw)

  run()
  shutdown()

  return


if __name__ == '__main__':
  main()



try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf
import numpy as np
import os

try:
  xavier_initializer = tf.contrib.layers.xavier_initializer()
except Exception:
  xavier_initializer = None


def disable_gpu():
  os.environ["CUDA_VISIBLE_DEVICES"] = '-1'
  return


def var_shape(x):
  out = [k.value for k in x.get_shape()]
  assert all(isinstance(a, int) for a in out), "shape function assumes that shape is fully known"
  return out


def intprod(x):
  return int(np.prod(x))


def numel(x):
  n = intprod(var_shape(x))
  return n


def flat_grad(loss, var_list, grad_ys=None):
  grads = tf.gradients(loss, var_list, grad_ys)
  return tf.concat([tf.reshape(grad, [numel(v)]) for (v, grad) in zip(var_list, grads)], axis=0)


def fc_net(input, layers_sizes, activation, reuse=None,
           flatten=False):  # build fully connected network
  curr_tf = input
  for i, size in enumerate(layers_sizes):
    with tf.variable_scope(str(i), reuse=reuse):
      curr_tf = tf.layers.dense(inputs=curr_tf,
                                units=size,
                                kernel_initializer=xavier_initializer,
                                activation=activation if i < len(layers_sizes) - 1 else None)
  if flatten:
    assert layers_sizes[-1] == 1
    curr_tf = tf.reshape(curr_tf, [-1])

  return curr_tf


def copy(sess, src, dst):
  assert len(src) == len(dst)
  sess.run(list(map(lambda v: v[1].assign(v[0]), zip(src, dst))))
  return


def flat_grad(loss, var_list):
  grads = tf.gradients(loss, var_list)
  return tf.concat(axis=0,
                   values=[tf.reshape(grad, [numel(v)]) for (v, grad) in zip(var_list, grads)])


def calc_logp_gaussian(x_tf, mean_tf, std_tf):
  dim = tf.to_float(tf.shape(x_tf)[-1])

  if mean_tf is None:
    diff_tf = x_tf
  else:
    diff_tf = x_tf - mean_tf

  logp_tf = -0.5 * tf.reduce_sum(tf.square(diff_tf / std_tf), axis=-1)
  logp_tf += -0.5 * dim * np.log(2 * np.pi) - tf.reduce_sum(tf.log(std_tf), axis=-1)

  return logp_tf


def calc_bound_loss(x_tf, bound_min, bound_max):
  # penalty for violating bounds
  violation_min = tf.minimum(x_tf - bound_min, 0)
  violation_max = tf.maximum(x_tf - bound_max, 0)
  violation = tf.reduce_sum(tf.square(violation_min), axis=-1) + tf.reduce_sum(
      tf.square(violation_max), axis=-1)
  loss = 0.5 * tf.reduce_mean(violation)
  return loss


class SetFromFlat(object):

  def __init__(self, sess, var_list, dtype=tf.float32):
    assigns = []
    shapes = list(map(var_shape, var_list))
    total_size = np.sum([intprod(shape) for shape in shapes])

    self.sess = sess
    self.theta = tf.placeholder(dtype, [total_size])
    start = 0
    assigns = []

    for (shape, v) in zip(shapes, var_list):
      size = intprod(shape)
      assigns.append(tf.assign(v, tf.reshape(self.theta[start:start + size], shape)))
      start += size

    self.op = tf.group(*assigns)

    return

  def __call__(self, theta):
    self.sess.run(self.op, feed_dict={self.theta: theta})
    return


class GetFlat(object):

  def __init__(self, sess, var_list):
    self.sess = sess
    self.op = tf.concat(axis=0, values=[tf.reshape(v, [numel(v)]) for v in var_list])
    return

  def __call__(self):
    return self.sess.run(self.op)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Script to render videos of the Proximal Policy Gradient algorithm.

Command line:

  python3 -m agents.scripts.visualize \
      --logdir=/path/to/logdir/<time>-<config> --outdir=/path/to/outdir/
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import gym
from gym import wrappers
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

from . import tools
from . import utility


def _create_environment(config, outdir):
  """Constructor for an instance of the environment.

  Args:
    config: Object providing configurations via attributes.
    outdir: Directory to store videos in.

  Returns:
    Wrapped OpenAI Gym environment.
  """
  if isinstance(config.env, str):
    env = gym.make(config.env)
  else:
    env = config.env()
  # Ensure that the environment has the specification attribute set as expected
  # by the monitor wrapper.
  if not hasattr(env, 'spec'):
    setattr(env, 'spec', getattr(env, 'spec', None))
  if config.max_length:
    env = tools.wrappers.LimitDuration(env, config.max_length)
  env = gym.wrappers.Monitor(env, outdir, lambda unused_episode_number: True)
  env = tools.wrappers.RangeNormalize(env)
  env = tools.wrappers.ClipAction(env)
  env = tools.wrappers.ConvertTo32Bit(env)
  return env


def _define_loop(graph, eval_steps):
  """Create and configure an evaluation loop.

  Args:
    graph: Object providing graph elements via attributes.
    eval_steps: Number of evaluation steps per epoch.

  Returns:
    Loop object.
  """
  loop = tools.Loop(None, graph.step, graph.should_log, graph.do_report, graph.force_reset)
  loop.add_phase('eval',
                 graph.done,
                 graph.score,
                 graph.summary,
                 eval_steps,
                 report_every=eval_steps,
                 log_every=None,
                 checkpoint_every=None,
                 feed={graph.is_training: False})
  return loop


def visualize(logdir, outdir, num_agents, num_episodes, checkpoint=None, env_processes=True):
  """Recover checkpoint and render videos from it.

  Args:
    logdir: Logging directory of the trained algorithm.
    outdir: Directory to store rendered videos in.
    num_agents: Number of environments to simulate in parallel.
    num_episodes: Total number of episodes to simulate.
    checkpoint: Checkpoint name to load; defaults to most recent.
    env_processes: Whether to step environments in separate processes.
  """
  config = utility.load_config(logdir)
  with tf.device('/cpu:0'):
    batch_env = utility.define_batch_env(lambda: _create_environment(config, outdir), num_agents,
                                         env_processes)
    graph = utility.define_simulation_graph(batch_env, config.algorithm, config)
    total_steps = num_episodes * config.max_length
    loop = _define_loop(graph, total_steps)
  saver = utility.define_saver(exclude=(r'.*_temporary/.*', r'global_step'))
  sess_config = tf.ConfigProto(allow_soft_placement=True)
  sess_config.gpu_options.allow_growth = True
  with tf.Session(config=sess_config) as sess:
    utility.initialize_variables(sess, saver, config.logdir, checkpoint, resume=True)
    for unused_score in loop.run(sess, saver, total_steps):
      pass
  batch_env.close()


def main(_):
  """Load a trained algorithm and render videos."""
  utility.set_up_logging()
  if not FLAGS.logdir or not FLAGS.outdir:
    raise KeyError('You must specify logging and outdirs directories.')
  FLAGS.logdir = os.path.expanduser(FLAGS.logdir)
  FLAGS.outdir = os.path.expanduser(FLAGS.outdir)
  visualize(FLAGS.logdir, FLAGS.outdir, FLAGS.num_agents, FLAGS.num_episodes, FLAGS.checkpoint,
            FLAGS.env_processes)


if __name__ == '__main__':
  FLAGS = tf.app.flags.FLAGS
  tf.app.flags.DEFINE_string('logdir', None, 'Directory to the checkpoint of a training run.')
  tf.app.flags.DEFINE_string('outdir', None, 'Local directory for storing the monitoring outdir.')
  tf.app.flags.DEFINE_string('checkpoint', None,
                             'Checkpoint name to load; defaults to most recent.')
  tf.app.flags.DEFINE_integer('num_agents', 1, 'How many environments to step in parallel.')
  tf.app.flags.DEFINE_integer('num_episodes', 5, 'Minimum number of episodes to render.')
  tf.app.flags.DEFINE_boolean('env_processes', True,
                              'Step environments in separate processes to circumvent the GIL.')
  tf.app.run()
